from __future__ import print_function
#  Name   :fig_book
#
#          this class to manage bundle of fig_pages
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************
import wx
import sys
from wx import WXK_RETURN

from ifigure.mto.fig_obj import FigObj
from ifigure.mto.fig_page import FigPage
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.utils.cbook as cbook
import ifigure.events

try:
    import ifigure.mdsplus
    has_mdsscope = True
except:
    has_mdsscope = False


class FigBook(FigObj):
    _image_id = [-1]
    _image_load_done = False

    def __init__(self, *args, **kywds):
        from ifigure.ifigure_config import iFigureConfig
        self._open = False
        self._screen_size = (0, 0)
        self._lock_scale = False
        self._viewer_class = 'BookViewer'
        self._viewer_module = 'ifigure.widgets.book_viewer.BookViewer'
        self._screen_ratio_lock = None
        # none locking, other size ratio of canvas size
        self._keep_data_in_tree = not iFigureConfig(
        ).setting['delbook_on_windowclose']
        super(FigBook, self).__init__(*args, **kywds)

    @classmethod
    def isFigBook(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'book'

    @classmethod
    def can_have_child(self, child=None):
        if isinstance(child, FigPage):
            return True
        if has_mdsscope:
            from ifigure.mdsplus.fig_mdsdata import FigMdsData
            if isinstance(child, FigMdsData):
                return True
        return False

    @property
    def isOpen(self):
        return self._open

    def onResize(self, evt):
        for page in self.walk_page():
            page.onResize(evt)

    def set_open(self, value):
        self._open = value
        self._image_update = False

    def classimage(self):
        if FigBook._image_load_done is False:
            FigBook._image_id = self.load_classimage()
            FigBook._image_load_done = True

        if self._open:
            return FigBook._image_id[1]
        else:
            return FigBook._image_id[0]

    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'book.png')
        idx2 = cbook.LoadImageFile(path, 'open_book.png')
        return [idx1, idx2]

    def pv_kshortcut(self, key):
        '''
        define keyborad short cut on project viewer
        must return a method responding key
        '''
        if not self.isOpen:
            return
        if key == WXK_RETURN:
            return self.onShow
        return None

    def generate_artist(self):
        pass
        # book does not have artist

    def del_artist(self, artist=None, delall=False):
        if delall:
            alist = self._artists
        else:
            alist = [artist]
        if len(alist) is len(self._artists):
            for name, child in self.get_children():
                child.del_artist(delall=True)
        else:
            print("deleting book artist partially...!?")

    def add_page(self, name=None, before=None, *args, **kywds):
        '''
        add a new page and return i_page
        '''
#        if 'before' in kywds:
#            before = kywds['before']
#            del  kywds['before']
#        else:
#            before = None
        fig_page = FigPage(*args, **kywds)
        if name is None:
            name = self.get_next_name(fig_page.get_namebase())

        ipage = self.add_child(name, fig_page, before=before)
        ipage = self.i_page(fig_page)
        return ipage

    def del_page(self, f_page):
        f_page.destroy()

    def num_page(self):
        '''
         return number of page
        '''
        i = 0
        for name, child in self.get_children():
            try:
                if child.isFigPage():
                    i = i+1
            except Exception:
                pass
        return i

    def get_page(self, ipage):
        '''
         return fig_page of ipage
         count only fig_page objects ignoring
         other children
        '''
        i = 0
        for name, child in self.get_children():
            try:
                if child.isFigPage():
                    if ipage == i:
                        return child
                    else:
                        i = i+1
            except Exception:
                pass
        return None

    def i_page(self, page):
        '''
        return ipage of page. ipage is an index
        of child without counting other than figpage 
        '''
        i = 0
        for name, child in self.get_children():
            if isinstance(child, FigPage):
                if child == page:
                    return i
                i = i+1
        return -1

    def walk_page(self):
        i = 0
        for name, child in self.get_children():
            if isinstance(child, FigPage):
                yield child

    def save_data2(self, data):
        app = self.get_root_parent().app
        viewer = app.find_bookviewer(self)
        if viewer is not None:
            s = viewer.canvas.get_canvas_screen_size()
            self._screen_size = (s[0], s[1])

        # the first element is version code
        param = {"_screen_size": self._screen_size,
                 "_keep_data_in_tree": self._keep_data_in_tree,
                 "_screen_ratio_lock": self._screen_ratio_lock,
                 "_viewer_class": self._viewer_class,
                 "_viewer_module": self._viewer_module}
        data['FigBook'] = (1, {}, param)
        data = super(FigBook, self).save_data2(data)
        return data

    def load_data2(self, data):
        self.handle_loaded_figobj_data(data, 'FigBook')
        param = data['FigBook'][2]
        for key in param:
            setattr(self, key, param[key])
        super(FigBook, self).load_data2(data)

    def toggle_keep_data(self):
        self._keep_data_in_tree = not (self._keep_data_in_tree)

    def get_keep_data(self):
        return self._keep_data_in_tree
#
#  tree viewer menu
#

    def tree_viewer_menu(self):
        app = self.get_root_parent().app
        if app in app.viewers:
            t = '-'
        else:
            t = ''

        m = [('Add Page', self.onAddPage, None), ]
        if self.isOpen:
            m = m + \
                [('Close Book', self.onCloseBook, None), ]

        elif (self.num_page() != 0 and not self.isOpen):
            m = m + \
                [(t+'Open in Main Window', self.onOpenMViwer, None),
                 ('Open in BookViewer', self.onOpenViwer, None), ]
            if (self._viewer_class != 'BookViewer' and
                    self._viewer_class != 'MDSScope'):
                m.append(('Open in '+self._viewer_class,
                          self.onOpenDefViewer, None))
            if has_mdsscope:
                m.append(('Open in MDSScope', self.onOpenMdsScope, None))
        else:
            m = m + \
                [('-Open in BookViwer', self.onOpenViwer, None), ]
            if self._viewer_class != 'BookViewer':
                m.append(('-Open in '+self._viewer_class,
                          self.onOpenDefViewer, None))
            if has_mdsscope:
                m.append(('-Open in MDSScope', self.onOpenMdsScope, None))
        if self._keep_data_in_tree:
            m = m + \
                [('^Keep book when window closed', self.onToggleKeepData, None), ]
        else:
            m = m + \
                [('*Keep book when window closed', self.onToggleKeepData, None), ]
        m = m + [('---', None, None)] + \
            super(FigBook, self).tree_viewer_menu()
        return m

    def onToggleKeepData(self, e):
        self.toggle_keep_data()
        e.Skip()

    def onAddPage(self, e):
        self.add_page()
        e.Skip()

    def onShow(self, e):
        id = e.GetEventObject()
        ifigure.events.SendShowPageEvent(self, id)

    def onOpenDefViewer(self, e):
        mod = __import__(self._viewer_module)
        md = sys.modules[self._viewer_module]
        the_class = getattr(md, self._viewer_class)
        ifigure.events.SendOpenBookEvent(self,
                                         w=e.GetEventObject(),
                                         viewer=the_class)
        e.Skip()

    def onOpenMViwer(self, e):
        app = self.get_root_parent().app
        app.open_book_in_appwindow(self)

    def onOpenViwer(self, e):
        from ifigure.widgets.book_viewer import BookViewer
        ifigure.events.SendOpenBookEvent(self,
                                         w=e.GetEventObject(),
                                         viewer=BookViewer)
        e.Skip()

    def onOpenMdsScope(self, e):
        from ifigure.mdsplus.mdsscope import MDSScope
        ifigure.events.SendOpenBookEvent(self,
                                         w=e.GetEventObject(),
                                         viewer=MDSScope)
        e.Skip()

    def onCloseBook(self, e):
        viewer = self.get_root_parent().app.find_bookviewer(self)
        ovalue = self._keep_data_in_tree
        self._keep_data_in_tree = True

        if viewer is not None:
            if viewer is not wx.GetApp().GetTopWindow():
                viewer.onClose(None)
            else:
                viewer.close_figurebook()
        self._keep_data_in_tree = ovalue
        e.Skip()

    def get_keep_data_in_tree(self):
        return self._keep_data_in_tree

    def set_keep_data_in_tree(self, value):
        self._keep_data_in_tree = value

    def find_bookviewer(self):
        app = self.get_root_parent().app
        return app.find_bookviewer(self)

    from ifigure.widgets.at_wxthread import at_wxthread
    @at_wxthread
    def Open(self, vc):
        app = wx.GetApp().TopWindow
        if self.isOpen:
            v = app.find_bookviewer(self)
            if isinstance(v, vc):
                return v
            v.onClose()
        viewer = app.open_book_in_newviewer(vc, self, ipage=0)
        return viewer

    @at_wxthread
    def Close(self):
        app = wx.GetApp().TopWindow
        if self.isOpen:
            v = app.find_bookviewer(self)
            v.onClose()



from __future__ import print_function

"""

  class:FigureCanvasWxMod

      A subclass of FigureCanvas to draw figures managed
      by figobj efficiently. This class provides...

      1) off-screen bitmap data storage to update highlight
         and other user interaction without drawing whole
         figure
      2) bitmap buffer for each axes to allow for updating
         one axes in figure screen without drawing all other
         axes


"""

# uncomment the following to use wx rather than wxagg
import matplotlib
import wx
from matplotlib.backends.backend_wx import FigureCanvasWx as Canvas
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as CanvasAgg
from matplotlib.backends.backend_wx import RendererWx
from ifigure.utils.cbook import EraseBitMap
from operator import itemgetter


class CanvasMod(object):
    '''
    CanvasMod provides modification to Wx and WxAgg
    '''

    def __init__(self, *args, **kargs):
        #        super(FigureCanvasWxMod, self).__init__(*args, **kargs)
        self.frame_bitmap = None
        self.others_bitmap = None
        self.figure_bitmap = None
        self.empty_bytes = []

    def draw_by_bitmap(self, *args, **kargs):
        print('draw bitmap')
        self.renderer = self.make_renderer()
#       gc = self.renderer.new_gc()
#       gc.select()
#        self.renderer.handle_clip_rectangle(gc)
        gc = None
        self.frame_bitmap = self.make_bitmap(self.figure.draw_frame, gc)
        self.make_axes_bitmap(gc)
        self.others_bitmap = self.make_bitmap(self.figure.draw_others, gc)
#        gc.unselect()

        self.erase_bitmap()
#        EraseBitMap(self.bitmap)

        bmplist = self.make_bmp_list()
        self.draw_from_bitmap(bmplist)
        self._isDrawn = True
        self.figure_bitmap = self.bitmap.GetSubBitmap(
            (0, 0,
             self.bitmap.GetWidth(),
             self.bitmap.GetHeight()))

    def draw(self, drawDC=None):
        self.draw_by_bitmap()
        self.gui_repaint(drawDC=drawDC)

#    def draw_all(self, drawDC=None):
#        pass

    def sorted_axes_list(self):
        # a list of (zorder, func_to_call, list_of_args)
        dsu = []
        for a in self.figure.axes:
            if (hasattr(a, 'figobj') and
                    a.figobj is not None):
                dsu.append((a.get_zorder(), a))
        dsu.sort(key=itemgetter(0))
        return dsu

    def make_bmp_list(self):
        bmplist = []
        bmplist.append((self.frame_bitmap, 0, 0))

        dsu = self.sorted_axes_list()
        for zorder, a in dsu:
            bmplist.append(a.figobj.get_bmp())
            # bmplist.append(a.figobj.get_bmp())
        bmplist.append((self.others_bitmap, 0, 0))
        return bmplist

    def erase_bitmap_by_copyfrombuffer(self):
        # erase bitmap by 0,0,0,0 to prepare
        # transparent empty screen
        import array
        bpp = 4  # bytes per pixel
        w = self.bitmap.GetWidth()
        h = self.bitmap.GetHeight()

        if (len(self.empty_bytes) !=
                self.bitmap.GetWidth()*self.bitmap.GetHeight()*bpp):
            self.empty_bytes = array.array('B', [0] * w*h*bpp)

        self.bitmap.CopyFromBufferRGBA(self.empty_bytes)

        return
#        self.renderer = RendererWx(self.bitmap, self.figure.dpi)
#        gc = self.renderer.new_gc()
#        gc.select()
#        gc.dc.SetBackground(wx.TRANSPARENT_BRUSH)
#        gc.dc.SetBackground(wx.TRANSPARENT_BRUSH)
#        gc.dc.Clear()
#        gc.unselect()
#        return
#        dc = wx.MemoryDC()
        dc = self.renderer.gc.dc
#        dc.SelectObject(self.bitmap)
        dc.SetBackground(wx.Brush(wx.Colour(0, 0, 0, 255)))
#        dc.SetBackground(wx.Brush('white'))
        dc.SetBackgroundMode(wx.SOLID)
        dc.SetLogicalFunction(wx.CLEAR)
        dc.Clear()
#        dc.SetBrush(wx.Brush(wx.Colour(125,0,0,125)))
        dc.SetLogicalFunction(wx.COPY)
#        dc.SetBackground(wx.Brush('white'))
#        dc.DrawRectangle(0,0, self.bitmap.GetWidth(),self.bitmap.GetHeight()/2)

#        dc.DrawRectangle(0,0, 100,100)
#        dc.SelectObject(wx.NullBitmap)
#        del dc
        # self.bitmap.SetMaskColour('black')

    def erase_bitmap(self):
        self.erase_bitmap_by_copyfrombuffer()
        return

    def make_bitmap(self, func, gc):
        '''
        prepare renderer.bitmap 
        '''
        self.erase_buffer()
#        gc = self.renderer.new_gc()
#        self.renderer.handle_clip_rectangle(gc)
#        gc.select()
        func(self.renderer)
#        gc.unselect()
        self.prepare_bitmap()
        bitmap = self.bitmap.GetSubBitmap(
            (0, 0,
             self.bitmap.GetWidth(),
             self.bitmap.GetHeight()))
        return bitmap  # self.make_true_bitmap_copy(bitmap)

    def make_axes_bitmap(self, gc):
        #        gc = self.renderer.new_gc()
        #        gc.select()
        dsu = self.sorted_axes_list()
        for zorder, a in dsu:
            if not a.figobj.get_bmp_update():
                self.erase_buffer()
                self.figure.draw_axes(self.renderer, a)
                self.prepare_bitmap()
                box = a.figobj.get_axesartist_extent(renderer=self.renderer)
                h = self.renderer.bitmap.GetHeight()
#               bmp =  self.make_subbitmap_copy(
#                            box[0], h-box[3],
#                            box[1]-box[0],
#                            box[3]-box[2])
                bmp = self.bitmap.GetSubBitmap(
                    (box[0], h-box[3],
                     box[1]-box[0],
                     box[3]-box[2]))
#               bmp = self.make_true_bitmap_copy(bmp)
                a.figobj.set_bmp(bmp, box[0], h-box[3])
#        gc.unselect()

    def draw_artist(self, drawDC=None, alist=None):
        """
        Render the figure using RendererWx instance renderer, or using a
        previously defined renderer if none is specified.
        """
        if alist is None:
            return
        if len(alist) == 0:
            return
        if drawDC is None:
            drawDC = wx.ClientDC(self)
        # print 'draw artist'
        self.erase_buffer()
        self.renderer = self.make_renderer()
        for a in alist:
            a.draw(self.renderer)
        self.prepare_bitmap()
        bitmap = self.bitmap.GetSubBitmap(
            (0, 0,
             self.bitmap.GetWidth(),
             self.bitmap.GetHeight()))
        self.erase_bitmap()
        self.canvas_draw_bitmap(self.figure_bitmap, 0, 0)
        self.canvas_draw_bitmap(bitmap, 0, 0)

        self.gui_repaint(drawDC=drawDC)

    def Copy_to_Clipboard_mod(self, event=None):
        "this is for debug purpose"
        "copy internal bitmaps of canvas to system clipboard"

        bmp_obj = wx.BitmapDataObject()
        bmp_obj.SetBitmap(self.figure_bitmap)
        wx.TheClipboard.Open()
        wx.TheClipboard.SetData(bmp_obj)
        wx.TheClipboard.Close()
        wx.TheClipboard.Flush()

    def make_true_bitmap_copy(self, bitmap):
        bmp = bitmap.GetSubBitmap(
            (0, 0,
             bitmap.GetWidth(),
             bitmap.GetHeight()))
        img = bmp.ConvertToImage()
        img.SaveFile('saved.png', wx.BITMAP_TYPE_PNG)
        return wx.BitmapFromImage(img)


class FigureCanvasWxModAgg(CanvasMod, CanvasAgg):
    def __init__(self, *args, **kargs):
        CanvasAgg.__init__(self, *args, **kargs)
        CanvasMod.__init__(self, *args, **kargs)

#    def _onSize(self, evt):
#        print '!!!! onSize'
#        super(FigureCanvasWxModAgg, self)._onSize(evt)

    def draw_all(self, drawDC=None):
        ''' 
        draw everything from scratch 
        mostly debugging purpose
        '''
        print('draw_all')
        self.figure.figobj.reset_axesbmp_update()
        CanvasAgg.draw(self)

    def erase_buffer(self):
        if hasattr(self, '_lastKey'):
            del self._lastKey
            self.renderer = self.make_renderer()

    def make_renderer(self):
        return self.get_renderer()

    def prepare_bitmap(self):
        from matplotlib.backends.backend_wxagg import _convert_agg_to_wx_bitmap
        self.bitmap = _convert_agg_to_wx_bitmap(self.get_renderer(), None)
        self.renderer.bitmap = self.bitmap

    def draw_from_bitmap(self, bmplist):
        if not self.figure.get_visible():
            return
        self.figure.draw_from_bitmap(self.renderer)
        dc = wx.MemoryDC()
        dc.SelectObject(self.bitmap)
        gfx_ctx = wx.GraphicsContext.Create(dc)
        for bmp, x, y in bmplist:
            gfx_ctx.DrawBitmap(bmp, x, y,
                               bmp.GetWidth(),
                               bmp.GetHeight())
        dc.SelectObject(wx.NullBitmap)

    def canvas_draw_bitmap(self, bitmap, x, y):
        dc = wx.MemoryDC()
        dc.SelectObject(self.bitmap)
        dc.DrawBitmap(bitmap, x, y)
        dc.SelectObject(wx.NullBitmap)

#    def _onSize(self, event):
#        if hasattr(self, "_on_size"):
#            return self._on_size(event)
#        else:
#            return super(FigureCanvasWxModAgg, Canvas)._onSize(event)


class FigureCanvasWxMod(CanvasMod, Canvas):
    def __init__(self, *args, **kargs):
        Canvas.__init__(self, *args, **kargs)
        CanvasMod.__init__(self, *args, **kargs)

    def draw_all(self, drawDC=None):
        ''' 
        draw everything from scratch 
        mostly debugging purpose
        '''
        print('draw_all')
        self.figure.figobj.reset_axesbmp_update()
        Canvas.draw(self)

    def make_renderer(self):
        return RendererWx(self.bitmap, self.figure.dpi)

    def prepare_bitmap(self):
        pass

    def draw_from_bitmap(self, bmplist):
        if not self.figure.get_visible():
            return
        self.figure.draw_from_bitmap(self.renderer)
        gc = self.renderer.new_gc()
        self.renderer.handle_clip_rectangle(gc)
        gc.select()
        for bmp, x, y in bmplist:
            gc.gfx_ctx.DrawBitmap(bmp, x, y,
                                  bmp.GetWidth(),
                                  bmp.GetHeight())
        gc.unselect()

    def erase_buffer(self):
        self.erase_bitmap()

    def canvas_draw_bitmap(self, bitmap, x, y):
        gc = self.renderer.new_gc()
        gc.select()
#        self.renderer.handle_clip_rectangle(gc)
        # gc.gfx_ctx.ResetClip()
        gc.gfx_ctx.DrawBitmap(bitmap, x, y,
                              bitmap.GetWidth(),
                              bitmap.GetHeight())
        gc.unselect()

#    def _onSize(self, event):
#        if hasattr(self, "_on_size"):
#            return self._on_size(event)
#        else:
#            return super(FigureCanvasWxMod, Canvas)._onSize(event)



from __future__ import print_function
import numpy as np
######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.
#      Strong recommendation : make module "independent".
#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  module loading
#      order-dependency problem.
#
#   name of module/class
module_name = 'sample_module'
class_name = 'sample_module'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
menu = [("MenuName", "onMenu", True)]
method = ['test', 'onMenu']

icon = 'cog.png'
######## Do not chage this field ####################
can_have_child = False
has_private_owndir = False
nosave_var0 = False
######################################################


def test(self):
    print(self)
    pass


def onMenu(self, e):
    print(self)
    print(e)
    pass



from __future__ import print_function
import six
import wx
import collections

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('iFigureCanvas')

from ifigure.utils.edit_list import TextDropTarget

color_map = {'g': 'green',
             'r': 'red',
             'b': 'blue',
             'k': 'black',
             'y': 'yellow',
             'm': 'magenta',
             'c': 'cyan'}
max_history = 15


class ShotNumberCtrl(wx.stc.StyledTextCtrl):
    def __init__(self, *args, **kargs):
        self._use_escape = True

        if not 'style' in kargs: kargs['style'] = 0
        kargs['style'] = kargs['style'] |  wx.TE_PROCESS_ENTER
        
        super(ShotNumberCtrl, self).__init__(*args, **kargs)
        
        self.Bind(wx.EVT_KEY_DOWN, self.onKeyPressed)
        self.Bind(wx.EVT_TEXT_ENTER, self.onEnter)
        self.Bind(wx.EVT_LEFT_DOWN, self.onDragInit)
        self.Bind(wx.stc.EVT_STC_CHANGE, self.onContentChanged)
        dt1 = TextDropTarget(self)
        self.SetDropTarget(dt1)
        min_w = 30
        txt_w, txt_h = self.Parent.GetTextExtent('A')
        self.SetMinSize((txt_w*20, txt_h+5))
        self.SetMaxSize((-1, txt_h+5))
        self.SetUseHorizontalScrollBar(False)
        self.SetUseVerticalScrollBar(False)
        self.SetReadOnly(False)
        for x in range(3):
            self.SetMarginWidth(x, 0)
        self._key_history_st1 = collections.deque([], maxlen=max_history)
        self._key_history_st2 = collections.deque([], maxlen=max_history)
        self._len_order = 5
        self._single_mode = False

    def set_color_style(self, order):
        self._len_order = len(order)
        for i, x in enumerate(order):
            c = color_map[x] if x in color_map else x
            if isinstance(c, tuple):
                c = [c[0]*255, c[1]*255, c[2]*255, c[3]*255]
            self.StyleSetForeground(i, c)
        self.StyleSetForeground(10, (0, 0, 0, 0))

    def set_style_bits(self):
        txt = str(self.GetValue())

        pos = self.GetCurrentPos()
        sel1, sel2 = self.GetSelection()
        arr = [x.split(',') for x in txt.split(';')]
        
        if txt.startswith('='):
            sarr = [['\x10'*len(xx) for k, xx in enumerate(x)] for x in arr]
        else:
            sarr = [[chr(k % self._len_order)*len(xx)
                     for k, xx in enumerate(x)] for x in arr]

        sarr = '\x10'.join(['\x10'.join(x) for x in sarr])
        bbb =''.join([''.join(x) for x in zip(txt, sarr)])
        
        self.ClearAll()
        
        #dprint1("bbb=", bbb.__repr__())

        if six.PY3:
            bbb = memoryview(bbb.encode('latin-1'))
        self.AddStyledText(bbb)
        self.SetCurrentPos(pos)
        self.SetSelection(sel1, sel2)

    def SetSingleMode(self, value):
        self._single_mode = value

    def SetValue(self, value):
        if self._single_mode:
            wx.stc.StyledTextCtrl.SetValue(
                self, (value.split(','))[0].split(';')[0])
        else:
            wx.stc.StyledTextCtrl.SetValue(self, value)

    def onContentChanged(self, evt):
        self.Unbind(wx.stc.EVT_STC_CHANGE)
        try:
            self.set_style_bits()
        except:
            import traceback
            traceback.print_exc()
            pass
        self.Bind(wx.stc.EVT_STC_CHANGE, self.onContentChanged)

    def onKeyPressed(self, event):
        key = event.GetKeyCode()
        if hasattr(event, 'RawControlDown'):
            controlDown = event.RawControlDown()
        else:
            controlDown = event.ControlDown()
        altDown = event.AltDown()

        if key == wx.WXK_UP:
            if len(self._key_history_st1) == 0:
                return
            v = self._key_history_st1.pop()
            self._key_history_st2.append(v)
            if len(self._key_history_st1) == 0:
                return
            self.SetValue(self._key_history_st1[-1])
            return
        elif key == wx.WXK_DOWN:
            if len(self._key_history_st2) == 0:
                return
            v = self._key_history_st2.pop()
            self._key_history_st1.append(v)
            if len(self._key_history_st1) == 0:
                return
            self.SetValue(self._key_history_st1[-1])
            return
        elif key == 67 and controlDown:  # ctrl + C (copy)
            self.Copy()
            return
        elif key == 87 and controlDown:  # ctrl + X (cut)
            self.Cut()
            return
        elif key == 88 and controlDown:  # ctrl + W (cut)
            self.Cut()
            return
        elif key == 86 and controlDown:  # ctrl + V (paste)
            self.Paste()
            return
        elif key == 89 and controlDown:  # ctrl + Y (paste)
            self.Paste()
            return
        elif key == 70 and controlDown:  # ctrl + F
            self.SetInsertionPoint(self.GetInsertionPoint()+1)
            return
        elif key == 66 and controlDown:  # ctrl + B
            self.SetInsertionPoint(self.GetInsertionPoint()-1)
            return
        elif key == 65 and controlDown:  # ctrl + A (beginning)
            print('move to front')
            self.SetInsertionPoint(0)
            self.SetSelection(0, 0)
            return
        elif key == 69 and controlDown:  # ctrl + E
            self.SetInsertionPoint(self.GetLastPosition())
            return
        elif key == 75 and controlDown:  # ctrl + K
            ### works only for single line ###
            self.SetSelection(self.GetInsertionPoint(),
                              self.GetLastPosition())
            self.Cut()
            return
        elif key == 44:  # ','
            if self._single_mode:
                return
        elif key == 59:  # ';'
            if self._single_mode:
                return
        elif key == wx.WXK_UP:
            if len(self._key_history_st1) == 0:
                return
            v = self._key_history_st1.pop()
            self._key_history_st2.append(v)
            if len(self._key_history_st1) == 0:
                return
            self.SetValue(self._key_history_st1[-1])
            return
        elif key == wx.WXK_DOWN:
            if len(self._key_history_st2) == 0:
                return
            v = self._key_history_st2.pop()
            self._key_history_st1.append(v)
            if len(self._key_history_st1) == 0:
                return
            self.SetValue(self._key_history_st1[-1])
            return

        if key == wx.WXK_RETURN:
            evt = wx.PyCommandEvent(wx.EVT_TEXT_ENTER.typeId,
                                    self.GetId())
            wx.PostEvent(self.GetEventHandler(), evt)
            return

        event.Skip()

    def onEnter(self, evt):
        v = self.GetValue()
        self._key_history_st1.append(v)
        evt.Skip()


#    def onEnter(self, evt):
#        self.GetParent().send_event(self, evt)

    def onDragInit(self, e):
        sel = self.GetStringSelection()
        if sel == '':
            e.Skip()
            return
        """ Begin a Drag Operation """
        # Create a Text Data Object, which holds the text that is to be dragged
        # app=wx.GetApp()
        p = self
        while p.GetParent() is not None:
            p = p.GetParent()

        p._text_clip = sel

        tdo = wx.TextDataObject(sel)
        tds = wx.DropSource(self)
        tds.SetData(tdo)
        tds.DoDragDrop(True)



from __future__ import print_function

""" MODULE images2gif

Provides a function (writeGif) to write animated gif from a series
of PIL images or numpy arrays.

This code is provided as is, and is free to use for all.

Almar Klein (June 2009)

- based on gifmaker (in the scripts folder of the source distribution of PIL)
- based on gif file structure as provided by wikipedia

"""

import glob

PIL = None
if PIL is None:
    try:
        import PIL
        from PIL import Image, ImageChops
        from PIL.GifImagePlugin import getheader, getdata
    except ImportError:
        PIL = None
# 2014 09 S. Shiraiwa
# import PIL is not right on some platform?
if PIL is None:
    try:
        import Image
        import ImageChops
        from GifImagePlugin import getheader, getdata
        PIL = Image
    except ImportError:
        PIL = None

try:
    import numpy as np
except ImportError:
    np = None

# getheader gives a 87a header and a color palette (two elements in a list).
# getdata()[0] gives the Image Descriptor up to (including) "LZW min code size".
# getdatas()[1:] is the image data itself in chuncks of 256 bytes (well
# technically the first byte says how many bytes follow, after which that
# amount (max 255) follows).


def intToBin(i):
    """ Integer to two bytes """
    # devide in two parts (bytes)
    i1 = i % 256
    i2 = int(i/256)
    # make string (little endian)
    return chr(i1) + chr(i2)


def getheaderAnim(im):
    """ Animation header. To replace the getheader()[0] """
    bb = "GIF89a"
    bb += intToBin(im.size[0])
    bb += intToBin(im.size[1])
    bb += "\x87\x00\x00"
    return bb


def getAppExt(loops=0):
    """ Application extention. Part that secifies amount of loops. 
    if loops is 0, if goes on infinitely.
    """
    bb = "\x21\xFF\x0B"  # application extension
    bb += "NETSCAPE2.0"
    bb += "\x03\x01"
    if loops == 0:
        loops = 2**16-1
    bb += intToBin(loops)
    bb += '\x00'  # end
    return bb


def getGraphicsControlExt(duration=0.1):
    """ Graphics Control Extension. A sort of header at the start of
    each image. Specifies transparancy and duration. """
    bb = '\x21\xF9\x04'
    bb += '\x08'  # no transparancy
    bb += intToBin(int(duration*100))  # in 100th of seconds
    bb += '\x00'  # no transparant color
    bb += '\x00'  # end
    return bb


def _writeGifToFile(fp, images, durations, loops):
    """ Given a set of images writes the bytes to the specified stream.
    """

    # init
    frames = 0
    previous = None

    for im in images:

        if not previous:
            # first image

            # gather data
            # This one stopped working
            # PIL might have changed??? 2016/03/30
            palette = getheader(im)[1]
            if palette is None:
                palette = getheader(im)[0][-1]

            data = getdata(im)
            imdes, data = data[0], data[1:]
            header = getheaderAnim(im)
            appext = getAppExt(loops)
            graphext = getGraphicsControlExt(durations[0])

            # write global header
            fp.write(header)
            fp.write(palette)
            fp.write(appext)

            # write image
            fp.write(graphext)
            fp.write(imdes)
            for d in data:
                fp.write(d)

        else:
            # gather info (compress difference)
            data = getdata(im)
            imdes, data = data[0], data[1:]
            graphext = getGraphicsControlExt(durations[frames])

            # write image
            fp.write(graphext)
            fp.write(imdes)
            for d in data:
                fp.write(d)

#             # delta frame - does not seem to work
#             delta = ImageChops.subtract_modulo(im, previous)
#             bbox = delta.getbbox()
#
#             if bbox:
#
#                 # gather info (compress difference)
#                 data = getdata(im.crop(bbox), offset = bbox[:2])
#                 imdes, data = data[0], data[1:]
#                 graphext = getGraphicsControlExt(durations[frames])
#
#                 # write image
#                 fp.write(graphext)
#                 fp.write(imdes)
#                 for d in data:
#                     fp.write(d)
#
#             else:
#                 # FIXME: what should we do in this case?
#                 pass

        # prepare for next round
        previous = im.copy()
        frames = frames + 1

    fp.write(";")  # end gif
    return frames


def writeGif(filename, images, duration=0.1, loops=0, dither=1):
    """ writeGif(filename, images, duration=0.1, loops=0, dither=1)
    Write an animated gif from the specified images. 
    images should be a list of numpy arrays of PIL images.
    Numpy images of type float should have pixels between 0 and 1.
    Numpy images of other types are expected to have values between 0 and 255.
    """

    if PIL is None:
        raise RuntimeError("Need PIL to write animated gif files.")

    images2 = []

    # convert to PIL
    for im in images:

        if isinstance(im, Image.Image):
            images2.append(im.convert('P', dither=dither))

        elif np and isinstance(im, np.ndarray):
            if im.dtype == np.uint8:
                pass
            elif im.dtype in [np.float32, np.float64]:
                im = (im*255).astype(np.uint8)
            else:
                im = im.astype(np.uint8)
            # convert
            if len(im.shape) == 3 and im.shape[2] == 3:
                im = Image.fromarray(im, 'RGB').convert('P', dither=dither)
            elif len(im.shape) == 2:
                im = Image.fromarray(im, 'L').convert('P', dither=dither)
            else:
                raise ValueError("Array has invalid shape to be an image.")
            images2.append(im)

        else:
            print(isinstance(im, Image.Image))
            raise ValueError("Unknown image type.")

    # check duration
    if hasattr(duration, '__len__'):
        if len(duration) == len(images2):
            durations = [d for d in duration]
        else:
            raise ValueError("len(duration) doesn't match amount of images.")
    else:
        durations = [duration for im in images2]

    # open file
    fp = open(filename, 'wb')

    # write
    try:
        n = _writeGifToFile(fp, images2, durations, loops)
        print((n, 'frames written'))
    finally:
        fp.close()


def glob2Gif(path, out, *args, **kwargs):
    """
    An extension to writeGif which takes a glob path instead of a list of
    images
    """

    files = glob.glob(path)
    images = []

    for f in sorted(files):
        im = Image.open(f)
        images.append(im)

    writeGif(out, images, *args, **kwargs)


if __name__ == '__main__':
    im = np.zeros((200, 200), dtype=np.uint8)
    im[10:30, :] = 100
    im[:, 80:120] = 255
    im[-50:-40, :] = 50

    images = [im*1.0, im*0.8, im*0.6, im*0.4, im*0]
    writeGif('lala3.gif', images, duration=0.5, dither=0)



#
#  Name   :section_editor
#
#          gui to edit section
#
#  History:
#          2012 spring v 0.1
#          2012 09 04 edge drag was added
#          2013 03    equal space and axis dge only
#          2014 02    bug fix.
#                     delete by edge drag
#               03    adding undo/redo
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu

__author__ = "Syun'ichi Shiraiwa"
__copyright__ = "Copyright, S. Shiraiwa, PiScope Project"
__credits__ = ["Syun'ichi Shiraiwa"]
__version__ = "0.5"
__maintainer__ = "S. Shiraiwa"
__email__ = "shiraiwa@psfc.mit.edu"
__status__ = "beta"

from ifigure.utils.wx3to4 import menu_AppendItem
from ifigure.widgets.undo_redo_history import UndoRedoAddRemoveArtists
from ifigure.widgets.undo_redo_history import UndoRedoGroupUngroupFigobj
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod
from ifigure.widgets.undo_redo_history import GlobalHistory
import wx
import weakref
import numpy as np

from ifigure.mto.fig_axes import FigAxes
#from ifigure.widgets.primitive_widgets import margin_widget, marginp_widget
from ifigure.widgets.margin_widget import MarginWidget, MarginpWidget
from ifigure.utils import geom as geom_util
import ifigure.events
#from matplotlib.figure import Figure
#from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as Canvas
from ifigure.utils.edit_list import EditListPanel, EDITLIST_CHANGED
from numpy.linalg import inv

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('LayoutEditor')


def _fix_event_data(w, h, event):
    event.xdata = float(event.x)/w
    event.ydata = float(event.y)/h
    if (event.xdata < 0) | (event.xdata > 1):
        event.xdata = None
    if (event.ydata < 0) | (event.ydata > 1):
        event.ydata = None
    return event


def send_area_request(canvas, axes=None, request=None, ac=None, name='area'):

    # canvas = ifigure_canvas
    fig_page = canvas._figure.figobj
    if ac is None:
        ac = []
    for mode, idx, value in request:
        if mode == 'm':  # modify
            fig_axes = fig_page.get_axes(idx)
            ac.append(UndoRedoFigobjMethod(fig_axes._artists[0],
                                           'area', value))
        elif mode == 'a':  # add
            iax = fig_page.add_axes(area=value)
            ax = fig_page.get_axes(iax)
            ax.realize()
            sel = [weakref.ref(ax._artists[0])]
            ac.append(UndoRedoAddRemoveArtists(artists=sel, mode=0))
        elif mode == 'd':  # delete
            fig_axes = fig_page.get_axes(idx)
            sel = [weakref.ref(a) for a in fig_axes._artists]
            ac.append(UndoRedoAddRemoveArtists(artists=sel, mode=1))

    window = canvas.GetTopLevelParent()
    GlobalHistory().get_history(window).make_entry(ac,
                                                   menu_name=name)


class AskRC(wx.Dialog):
    def __init__(self, parent, id, title):
        wx.Dialog.__init__(self, parent, id, title)

        box = wx.BoxSizer(wx.VERTICAL)

        from ifigure.utils.wx3to4 import FlexGridSizer
        vbox = FlexGridSizer(2, 2)

        stline1 = wx.StaticText(self, 3, 'R')
        self.sc1 = wx.SpinCtrl(self, -1, '')
        self.sc1.SetRange(1, 8)
        stline2 = wx.StaticText(self, 3, 'C')
        self.sc2 = wx.SpinCtrl(self, -1, '')
        self.sc2.SetRange(1, 8)
        vbox.Add(stline1, 1,  wx.ALIGN_CENTER | wx.TOP, 10)
        vbox.Add(self.sc1, 1, wx.ALIGN_CENTER | wx.TOP, 10)
        vbox.Add(stline2, 1, wx.ALIGN_CENTER | wx.TOP, 10)
        vbox.Add(self.sc2, 1, wx.ALIGN_CENTER | wx.TOP, 10)

        sizer = self.CreateButtonSizer(wx.OK | wx.CANCEL)
        box.Add(vbox,  0, wx.ALIGN_CENTER | wx.ALL, 10)
        box.Add(sizer, 0, wx.ALIGN_CENTER | wx.ALL, 10)
        self.SetSizer(box)
        self.Fit()
        #self.Bind(wx.EVT_BUTTON, self.OnOK, id=wx.ID_OK)
        #self.Bind(wx.EVT_BUTTON, self.OnCANCEL, id=wx.ID_CANCEL)
        self.Center()

    def OnOK(self, event):
        self.Close()

    def OnCANCEL(self, event):
        self.Close()

    def GetValue(self):
        return (self.sc1.GetValue(), self.sc2.GetValue())


class layout_editor_popup(wx.Menu):
    def __init__(self, parent):
        super(layout_editor_popup, self).__init__(
            style=wx.WS_EX_PROCESS_UI_UPDATES)
        self.parent = weakref.ref(parent)
        self.mmi = []
        m = ["Split", "Merge", "Swap", "Distribute Horizontally",
             "Distribute Vertically",
             "Sort by Column", "Sort by Row",
             "Common Bottom Axis",
             "Common Left Axis", "Reset Common Axis", ]

        h = [self.onSplit, self.onMerge, self.onSwap, self.onDistH,
             self.onDistV, self.onSortCol, self.onSortRow,
             self.onCommBtm, self.onCommLft,
             self.onResetCommon, ]

        for k in range(0, len(m)):
            key = m[k]
            if key == "---":
                self.AppendSeparator()
                continue
            mmi = wx.MenuItem(self, wx.ID_ANY, key)
            menu_AppendItem(self, mmi)
            self.Bind(wx.EVT_MENU, h[k], mmi)
            self.mmi.append(mmi)

    def onSplit(self, e):
        dialog = AskRC(self.parent(), wx.ID_ANY, "Split Subplot")
        le = self.parent().layout_editor
        if dialog.ShowModal() == wx.ID_OK:
            rc = dialog.GetValue()
            area = le.area_hit
            new_areas = le.split_area(area, rc[0], rc[1])
#           self.parent.replace_area(area, new_areas)
            request = [('m', le.area.index(le.area_hit), new_areas[0])]
            for a in new_areas[1:]:
                request.append(('a', None, a))
            le.SetCanvasValue(request=request)
            le.area_hit = None
            le.SetEditorValue()
        dialog.Destroy()

    def onMerge(self, e):
        self.parent().layout_editor._wait4merge = True
        self.parent().layout_editor._merge_a = self.parent().layout_editor.area_hit

    def onSwap(self, e):
        self.parent().layout_editor._wait4swap = True
        self.parent().layout_editor._swap_a = self.parent().layout_editor.area_hit

    def onDistH(self, e):
        self.parent().layout_editor.distH(self.parent().layout_editor.area_hit)

    def onDistV(self, e):
        self.parent().layout_editor.distV(self.parent().layout_editor.area_hit)

    def onCommBtm(self, e):
        self.parent().layout_editor.apply_bottom_only(
            self.parent().layout_editor.area_hit)

    def onCommLft(self, e):
        self.parent().layout_editor.apply_left_only(
            self.parent().layout_editor.area_hit)

    def onSortRow(self, e):
        from ifigure.events import SendChangedEvent
        self.parent()._figure.figobj.sort_axes_row()
        self.parent().layout_editor.draw()
        fig = self.parent()._figure.figobj
        SendChangedEvent(fig, w=None, useProcessEvent=False)

    def onSortCol(self, e):
        from ifigure.events import SendChangedEvent
        self.parent()._figure.figobj.sort_axes_col()
        self.parent().layout_editor.draw()
        fig = self.parent()._figure.figobj
        SendChangedEvent(fig, w=None, useProcessEvent=False)

    def onResetCommon(self, evt):
        fig = self.parent()._figure
        f_page = fig.figobj
        ac = []
        request = []
        for f_axes in f_page.walk_axes():
            ac.append(UndoRedoFigobjMethod(f_axes._artists[0],
                                           'edge_only', [False]*4))
            ac.append(UndoRedoFigobjMethod(f_axes._artists[0],
                                           'extra_margin', [0, 0, 0, 0]))
            request.append(('m', f_page.get_iaxes(f_axes), f_axes.get_area()))

        # finish up
        send_area_request(self.parent(), request=request,
                          ac=ac, name='reset common axis')

    def update_ui(self):
        if self.parent().axes_selection() is None:
            return False
        if self.parent().axes_selection().figobj is None:
            return False
        if any(self.parent().axes_selection().figobj.get_edge_only()):
            self.mmi[-1].Enable(True)
        else:
            self.mmi[-1].Enable(False)
        return True
#       wx.CallAfter(self.SetEditorValue)


class layout_editor(object):
    def __init__(self, parent):
        self.parent = weakref.ref(parent)
        self.area = []
        self.rect = []
        self.fig_hl = []  # used to draw area
        self.fig_txt = []  # used to draw area
        self.fig_hl2 = []  # used to draw area during drag
        self.fig_hlp = []  # patches
        self.edge_hl = None
        self.edge_hit = []
        self._mid = None
        self.drag = 0

        self.use_def_margin = []
        self.area_hit = None
        self._wait4merge = False
        self._merge_a = None
        self._wait4swap = False
        self._swap_a = None
        self._disth = False
        self._distv = False
        self.page_margin = [0, 0, 0, 0]
        self.nomargin = False
        self._popup = layout_editor_popup(parent)
#       self. _load_canvas_value()

    @property
    def canvas(self):
        return self.parent()

    @property
    def mpl_canvas(self):
        return self.parent().canvas

    @property
    def figure(self):
        return self.parent()._figure

    def enter_layout_mode(self):
        self. _load_canvas_value()
        self.draw()

    def exit_layout_mode(self):
        if self._mid is not None:
            self.mpl_canvas.mpl_disconnect(self._mid)
        self._mid = None
        self.remove_artists()

    def remove_artists(self, drag=True, layout=True):
        if layout:
            for l in self.fig_hl:
                self.figure.lines.remove(l)
            self.fig_hl = []
            for l in self.fig_txt:
                self.figure.texts.remove(l)
            self.fig_txt = []
            for l in self.fig_hlp:
                self.figure.patches.remove(l)
            self.fig_hlp = []
        if drag:
            for l in self.fig_hl2:
                self.figure.lines.remove(l)

            self.fig_hl2 = []

    def replace_area(self, area, new_areas):
        idx = self.area.index(area)
        self.area.remove(area)
        self.area.insert(idx, new_areas[0])

        if len(new_areas) > 1:
            for k in range(1, len(new_areas)):
                self.area.append(new_areas[k])
        self.draw()

#   def onDraw(self, event):
#       dprint1('onDraw')
        # this is for  draw_event of mpl
    def draw_all(self):
        self._load_canvas_value()
        self.draw()

    def draw(self):
        from matplotlib.lines import Line2D
        from matplotlib.patches import Rectangle
        dprint2('draw')

        if self.nomargin:
            pm = [0, 0, 0, 0]
        else:
            pm = self.page_margin
  #      if self.parent.get_canvas() is None: return
        fig = self.figure
        f_page = fig.figobj

        self.remove_artists()

        rect1s = [(pm[0], pm[3], 1.-pm[1]-pm[0], 1.-pm[2]-pm[3])]
        rect2s = []

        for f_axes in f_page.walk_axes():
            area = f_axes.getp("area")
            rect1 = [(1.-pm[0]-pm[1])*area[0]+pm[0],
                     (1.-pm[2]-pm[3])*area[1]+pm[3],
                     area[2]*(1.-pm[0]-pm[1]),
                     area[3]*(1.-pm[2]-pm[3]), ]

            rect1s.append(rect1)
            if area == self.area_hit:
                rect2, void, void2 = f_axes.calc_rect()
                rect2s.append(rect2)
        for num, rect1 in enumerate(rect1s):
            x, y = self._area2xy(rect1)
            w, h = self.mpl_canvas.get_bitmap_width_height()
            x1 = (np.array(x)*(w-2)+1)/w
            y1 = (np.array(y)*(h-2)+1)/h
            hl = Line2D(x1, y1, marker=None,
                        color='r', linestyle='-',
                        markerfacecolor='k',
                        transform=self.figure.transFigure,
                        figure=self.figure, alpha=0.3,
                        linewidth=5)
            self.fig_hl.append(hl)
            self.figure.lines.append(hl)
            if num != 0:
                t = self.figure.text((max(x1)+min(x1))/2,
                                     (max(y1)+min(y1))/2,
                                     str(num-1),
                                     transform=self.figure.transFigure,
                                     color='red', size=18)
                self.fig_txt.append(t)
        for rect2 in rect2s:
            x, y = self._area2xy(rect2)
            x1 = (np.array(x)*(w-2)+1)/w
            y1 = (np.array(y)*(h-2)+1)/h
            hl = Rectangle((min(x1), min(y1)),
                           max(x1)-min(x1), max(y1)-min(y1),
                           facecolor='r', alpha=0.3,
                           transform=self.figure.transFigure,
                           figure=self.figure)
            self.fig_hlp.append(hl)
            self.figure.patches.append(hl)

        f_page = self.canvas._figure.figobj
        nsec = f_page.num_axes()
        nsec2 = f_page.num_axes(include_floating=True)
        for i in range(nsec2-nsec):
            f_axes = f_page.get_axes(iax=i+nsec)
            bb = f_axes._artists[0].get_window_extent().bounds
            x, y = self.figure.transFigure.inverted().transform((bb[0] + bb[2]/2.,
                                                                 bb[1] + bb[3]/2.))
            t = self.figure.text(x, y,
                                 str(i+nsec),
                                 color='green',
                                 size=18)
            self.fig_txt.append(t)

        self.canvas.draw_artist(self.fig_hl+self.fig_hlp + self.fig_txt)
        self.remove_artists()

    def draw_request(self, request):
        from matplotlib.lines import Line2D
        from matplotlib.patches import Rectangle

        if self.nomargin:
            pm = [0, 0, 0, 0]
        else:
            pm = self.page_margin
#       if self.parent.get_canvas() is None: return
        fig = self.figure
        f_page = fig.figobj
        self.remove_artists()  # layout=False)
#       for type, idx, area in request:
        areas = [area for type, idx, area in request]
        areas.append([0, 0, 1, 1])
        for area in areas:
            if area is None:
                continue
            rect1 = [(1.-pm[0]-pm[1])*area[0]+pm[0],
                     (1.-pm[2]-pm[3])*area[1]+pm[3],
                     area[2]*(1.-pm[0]-pm[1]),
                     area[3]*(1.-pm[2]-pm[3]), ]
            x, y = self._area2xy(rect1)
            w, h = self.mpl_canvas.get_bitmap_width_height()
            x1 = (np.array(x)*(w-2)+1)/w
            y1 = (np.array(y)*(h-2)+1)/h
            hl = Line2D(x1, y1, marker=None,
                        color='r', linestyle='-',
                        markerfacecolor='k',
                        transform=self.figure.transFigure,
                        figure=self.figure, alpha=0.3, linewidth=5)
            self.fig_hl2.append(hl)
            self.figure.lines.append(hl)
        self.canvas.draw_artist(self.fig_hl2)
        self.remove_artists()

    def split_area(self, area, r, c):
        w = float(area[2])/c
        h = float(area[3])/r
        ret = []
        for j in range(0, c):
            for i in range(0, r):
                ret.append([area[0]+j*w, area[1]+(r-i-1)*h, w, h])
        return ret

    def merge_area(self, a1, a2):
        box1 = self._area2box(a1)
        box2 = self._area2box(a2)

        new_box = None
        if (abs(box1[0] - box2[0]) < 0.0001 and
                abs(box1[2] - box2[2]) < 0.0001):
            # vertical merge
            if abs(box1[1] - box2[3]) < 0.001:
                new_box = [box1[0], min([box1[3], box2[1]]),
                           box2[2], max([box1[3], box2[1]])]
            if abs(box2[1] - box1[3]) < 0.001:
                new_box = [box1[0], min([box1[1], box2[3]]),
                           box2[2], max([box1[1], box2[3]])]
        if (abs(box1[1] - box2[1]) < 0.001 and
                abs(box1[3] - box2[3]) < 0.001):
            # horizontal merge
            if abs(box1[0] - box2[2]) < 0.001:
                new_box = [min([box1[2], box2[0]]), box1[1],
                           max([box1[2], box2[0]]), box2[3]]
            if abs(box2[0] - box1[2]) < 0.001:
                new_box = [min([box1[0], box2[2]]), box1[1],
                           max([box1[0], box2[2]]), box2[3]]

        if new_box is None:
            return None
        return self._box2area(new_box)

    def buttonpress(self, event):
        self. _load_canvas_value()
        # edge hit check
#       h = self.figure.get_figheight()*self.figure.get_dpi()
#       w = self.figure.get_figwidth()*self.figure.get_dpi()
        w, h = self.mpl_canvas.get_bitmap_width_height()
        event = _fix_event_data(w, h, event)

        self.edge_hit = []
        self.edge_hit2 = []
        self.edge_hit3 = ()
        self.drag = 0
        for i, area in enumerate(self.area):
            # edge hit overwrite event.x or event.y if it hits
            hit = self.check_edge_hit(area, event, w, h)
            if hit != -1:
                self.edge_hit.append([i, hit, False])
        for i, area in enumerate(self.area):
            hit2 = self.check_edge_hit2(area, event.x, event.y,
                                        w, h)
            if hit2 != -1:
                self.edge_hit2.append([i, hit2, False])
        if len(self.edge_hit) != 0:
            if event.key == 'shift':
                single = True
            else:
                single = False
            self.edge_hit3 = self.select_connected_edge(single=single)
            if self.edge_hit3[0] != -1:
                if self._mid is None:
                    self._last_request = None
                    self._mid = self.mpl_canvas.mpl_connect(
                        'motion_notify_event', self.motion)
                self.add_edge_hl()
                return
            else:
                self.edge_hit = []
                self.edge_hit2 = []
                self.edge_hit3 = ()

        # area hit check
        if self.nomargin:
            pm = [0, 0, 0, 0]
        else:
            pm = self.page_margin

        for area in self.area:
            rect1 = [(1.-pm[0]-pm[1])*area[0]+pm[0],
                     (1.-pm[2]-pm[3])*area[1]+pm[3],
                     area[2]*(1.-pm[0]-pm[1]),
                     area[3]*(1.-pm[2]-pm[3]), ]

            if geom_util.check_inside(rect1, event.xdata,
                                      event.ydata):
                self.area_hit = area
#             self.draw()
                self.SetEditorValue()
                return
        self.area_hit = self.area[0]
        self.SetEditorValue()

    def buttonrelease(self, event):
        w, h = self.mpl_canvas.get_bitmap_width_height()
        event = _fix_event_data(w, h, event)
        # cleaning...
        if self._mid is not None:
            self.mpl_canvas.mpl_disconnect(self._mid)
        self._mid = None
        self.remove_artists(layout=False)

        if event.button == 3:
            if len(self.edge_hit) == 0:
                # context menu
                #              scr_size=self.mpl_canvas.GetClientSize()
                if not self._popup.update_ui():
                    return
                self.mpl_canvas.PopupMenu(self._popup,
                                          [event.x, h-event.y])
            else:
                self.rm_edge_hl()
        if event.button == 1:
            #         print self._wait4merge
            if self._wait4merge:
                if self._merge_a is None:
                    return
                if self.area_hit is None:
                    return
                new_area = self.merge_area(self._merge_a, self.area_hit)
                if new_area is not None:
                    request = [('m', self.area.index(self._merge_a), new_area),
                               ('d', self.area.index(self.area_hit), None)]
                    self.SetCanvasValue(request=request)
                wx.CallAfter(self.SetEditorValue)
#               self.SetEditorValue()
#               self.canvas.draw()
                self._merge_a = None
                self._wait4merge = False
            if self._wait4swap:
                if self._swap_a is None:
                    return
                if self.area_hit is None:
                    return
                request = [('m', self.area.index(self.area_hit), self._swap_a),
                           ('m', self.area.index(self._swap_a), self.area_hit)]
                self.SetCanvasValue(request=request)
                wx.CallAfter(self.SetEditorValue)
#            self.canvas.draw()

                self._swap_a = None
                self._wait4swap = False

            if (self.drag == 1 and
                    len(self.edge_hit) != 0):
                self.rm_edge_hl()
                request = self.process_edge_drag(event)
                if request is not None:
                    self.SetCanvasValue(request=request)
#             wx.CallAfter(self.SetEditorValue)

            fig = self.figure
            f_page = fig.figobj

            ax = None
            for f_axes in f_page.walk_axes():
                if not isinstance(f_axes, FigAxes):
                    continue
                area = f_axes.getp("area")
                if area == self.area_hit:
                    ax = f_axes
                    break
            if ax is not None:
                ifigure.events.SendSelectionEvent(
                    f_axes, self.canvas.GetTopLevelParent(), [])

        # print self.area_hit

    def onpick(self, event):
        pass

    def motion(self, event):
        w, h = self.mpl_canvas.get_bitmap_width_height()
        event = _fix_event_data(w, h, event)

        if event.xdata is None or event.ydata is None:
            return
        self.drag = 1
        if (self.edge_hl is not None and
                len(self.edge_hit3) != 0):
            area_bk = [a[:] for a in self.area]
            request = self.process_edge_drag(event)
            self.area = area_bk
        if request is not None:
            self.draw_request(request)

    def add_edge_hl(self):
        self.edge_hl = True

    def rm_edge_hl(self):
        #       if self.edge_hl in self.axes.lines:
        #           self.edge_hl.remove()
        self.edge_hl = None

    def process_release_outside(self, event):
        cx, cy = self.canvas.GetClientSize()
        # rarea -> removed
        # earea -> expanded
        request = []
        irarea = []
        earea = None
        if event.x < 0:
            irarea = [k[0] for k in self.edge_hit2 if k[1] == 2]
            earea = [(k[0], self.area[k[0]])
                     for k in self.edge_hit2 if k[1] == 0]
            for idx, x in earea:
                x[2] = x[2]+x[0]
                x[0] = 0
                request.append(('m', idx, x))
        elif event.x > cx:
            irarea = [k[0] for k in self.edge_hit2 if k[1] == 0]
            earea = [(k[0], self.area[k[0]])
                     for k in self.edge_hit2 if k[1] == 2]
            for idx, x in earea:
                x[2] = 1.-x[0]
                request.append(('m', idx, x))

        elif event.y < 0:
            irarea = [k[0] for k in self.edge_hit2 if k[1] == 3]
            earea = [(k[0], self.area[k[0]])
                     for k in self.edge_hit2 if k[1] == 1]
            for idx, x in earea:
                x[3] = x[3]+x[1]
                x[1] = 0
                request.append(('m', idx, x))

        elif event.y > cy:
            irarea = [k[0] for k in self.edge_hit2 if k[1] == 1]
            earea = [(k[0], self.area[k[0]])
                     for k in self.edge_hit2 if k[1] == 3]
            for idx, x in earea:
                x[3] = 1 - x[1]
                request.append(('m', idx, x))

        if earea is not None and len(earea) == 0:
            # filling void
            if len(irarea) == 1:
                k = irarea[0]
                left, right, up, down = self._search_nearby(self.area[k])
                if len(left) > 0 and len(right) == 0:
                    for idx, x in left:
                        x[2] = 1 - x[0]
                        request.append(('m', idx, x))
                elif len(left) > 0 and len(right) > 0:
                    mid = self.area[k][0] + self.area[k][2]/2.0
                    for idx, x in right:
                        x[2] = x[0] + x[2] - mid
                        x[0] = mid
                        request.append(('m', idx, x))
                    for idx, x in left:
                        x[2] = mid - x[0]
                        x[0] = x[0]
                        request.append(('m', idx, x))
                elif len(left) == 0 and len(right) > 0:
                    for idx, x in right:
                        x[2] = x[0] + x[2]
                        x[0] = 0
                        request.append(('m', idx, x))
                elif len(up) > 0 and len(down) == 0:
                    for idx, x in up:
                        x[3] = x[1] + x[1]
                        x[1] = 0
                        request.append(('m', idx, x))
                elif len(up) > 0 and len(down) > 0:
                    mid = self.area[k][1] + self.area[k][3]/2.0
                    for idx, x in down:
                        x[3] = mid - x[1]
                        x[1] = x[1]
                        request.append(('m', idx, x))
                    for idx, x in up:
                        x[3] = x[1] + x[3] - mid
                        x[1] = mid
                        request.append(('m', idx, x))
                elif len(up) == 0 and len(down) > 0:
                    for idx, x in down:
                        x[3] = 1 - x[1]
                        request.append(('m', idx, x))

        for k in irarea:
            request.append(('d', k, None))

        if len(request) > 0:
            return request

    def process_edge_drag(self, event):
        def calc_push(areas, nextones, d, mode, min_d):
            if (mode == 0) | (mode == 2):
                dd = sum([areas[k][2]-min_d for k in nextones])
                ddd = [d*(areas[k][2]-min_d)/dd for k in nextones]
            else:
                dd = sum([areas[k][3]-min_d for k in nextones])
                ddd = [d*(areas[k][3]-min_d)/dd for k in nextones]

            return ([sum(ddd[k:]) for k in range(len(nextones))],
                    [sum(ddd[k+1:]) for k in range(len(nextones))])
#       event = self._fix_event_data(event)
        cx, cy = self.canvas.GetClientSize()
        if event.xdata is None or event.ydata is None:
            return self.process_release_outside(event)

        if self.nomargin:
            pm = [0., 0., 0., 0.]
        else:
            pm = self.page_margin

        # this maps events to normal coordinate so that
        # area can use it.
        event.xdata = (event.xdata - pm[0])/(1.-pm[0]-pm[1])
        event.ydata = (event.ydata - pm[3])/(1.-pm[2]-pm[3])
        event.x = (event.x - cx*pm[0])/(1.-pm[0]-pm[1])
        event.y = (event.y - cy*pm[3])/(1.-pm[2]-pm[3])

        if event.xdata > 1 or event.xdata < 0:
            event.xdata = None
        if event.ydata > 1 or event.ydata < 0:
            event.ydata = None

        if event.xdata is None or event.ydata is None:
            return self.process_release_outside(event)

        area_bk = [a[:] for a in self.area]
        # backup

        request = []
        min_dx = 10./cx
        min_dy = 10./cy
        ohit2 = [tmp for tmp in self.edge_hit2]  # original data
        for hit in self.edge_hit2:
            if not hit[2]:
                continue
            nextones = self.find_nexeones(hit[0], hit[1])

            if hit[1] == 0:
                x = self.area[hit[0]][0]+self.area[hit[0]][2]
                d = x - event.xdata
                if d < min_dx:
                    if len(nextones) == 0:
                        return self._last_request
#                       self.area = area_bk
#                       return
                    x = min_dx - d
                    x1, x2 = calc_push(self.area, nextones, x, 0, min_dx)
                    for i, k2 in enumerate(nextones):
                        self.push_area(k2, 0, x1[i], x2[i])
                    d = min_dx
                self.area[hit[0]][0] = event.xdata
                self.area[hit[0]][2] = d
            elif hit[1] == 1:
                y = self.area[hit[0]][1]+self.area[hit[0]][3]
                d = y - event.ydata
                if d < min_dy:
                    if len(nextones) == 0:
                        return self._last_request
                #           self.area = area_bk
#                       return
                    x = min_dy - d
                    x1, x2 = calc_push(self.area, nextones, x, 1, min_dy)
                    for i, k2 in enumerate(nextones):
                        self.push_area(k2, 1, x1[i], x2[i])
                    d = min_dy
                self.area[hit[0]][1] = event.ydata
                self.area[hit[0]][3] = d

            elif hit[1] == 2:
                d = event.xdata - self.area[hit[0]][0]
                if d < min_dx:
                    if len(nextones) == 0:
                        return self._last_request
#                       self.area = area_bk
#                       return
                    x = min_dx - d
                    x1, x2 = calc_push(self.area, nextones, x, 2, min_dx)
                    for i, k2 in enumerate(nextones):
                        self.push_area(k2, 2, -x1[i], -x2[i])
                    d = min_dx
  #                 self.area[hit[0]][0] = self.area[hit[0]][0] - min_d
                    self.area[hit[0]][0] = event.xdata - d
                self.area[hit[0]][2] = d

            elif hit[1] == 3:
                d = event.ydata - self.area[hit[0]][1]
                if d < min_dy:
                    if len(nextones) == 0:
                        return self._last_request
#                       self.area = area_bk
#                       return
                    x = min_dy - d
                    x1, x2 = calc_push(self.area, nextones, x, 3, min_dy)
                    for i, k2 in enumerate(nextones):
                        self.push_area(k2, 3, -x1[i], -x2[i])
                    d = min_dy
#                  self.area[hit[0]][1] = self.area[hit[0]][1] - min_d
                    self.area[hit[0]][1] = event.ydata - d
                self.area[hit[0]][3] = d

        for k, a in enumerate(self.area):
            request.append(('m', k, self.area[k]))

#       for mode, idx, area in request:
#           if area[3] < 0:
#                print self._search_connecting_area(self.area[idx], False)
        d = self.edge_hit3[2] - self.edge_hit3[1]
        if self.edge_hit3[4] == 1:  # boundary drag mode..
            for hit in ohit2:
                if not hit:
                    continue
                if hit[1] == 0:
                    newa = [0, self.edge_hit3[1], event.xdata, d]
                    break
                if hit[1] == 1:
                    newa = [self.edge_hit3[1], 0, d, event.ydata]
                    break
                if hit[1] == 2:
                    newa = [event.xdata, self.edge_hit3[1],
                            abs(1-event.xdata), d]
                    break
                if hit[1] == 3:
                    newa = [self.edge_hit3[1],
                            event.ydata, d, abs(1-event.ydata)]
                    break
            request.append(('a', None, newa))
#          self.area.append(newa)
        for mode, idx, a in request:
            if (a[2]*cx < 10) | (a[3]*cy < 10):
                return self._last_request
#                self.area = area_bk
                return
        if abs(sum([abs(a[2]*a[3]) for mode, idx, a in request]) - 1) > 0.01:
            return self._last_request
#           self.area = area_bk
        else:
            self._last_request = request
            return request

    def push_area(self, k, mode, d, d2):
        # d move of the first side
        # d2 move of the other side
        if mode == 0:
            self.area[k][0] = self.area[k][0] + d
            self.area[k][2] = self.area[k][2] - d + d2
        elif mode == 1:
            self.area[k][1] = self.area[k][1] + d
            self.area[k][3] = self.area[k][3] - d + d2
        elif mode == 2:
            self.area[k][0] = self.area[k][0] + d2
            self.area[k][2] = self.area[k][2] + d - d2
        elif mode == 3:
            self.area[k][1] = self.area[k][1] + d2
            self.area[k][3] = self.area[k][3] + d - d2

    def find_nexeones(self, k, mode, ans=None):
        # find next one:
        def check_dir(a1, a2, mode):
            #  mode = 0 : true if a1 is left of a2
            if mode == 0:
                return a1[0] < a2[0]
            elif mode == 1:
                return a1[1] < a2[1]
            elif mode == 2:
                return a1[0] > a2[0]
            elif mode == 3:
                return a1[1] > a2[1]

        def op_mode(mode):
            if mode == 0:
                return 2
            elif mode == 1:
                return 3
            elif mode == 2:
                return 0
            elif mode == 3:
                return 1

        h = True if (mode == 0) | (mode == 2) else False
        omode = op_mode(mode)
        if ans is None:
            ans = []
        for k2, a in enumerate(self.area):
            if (self._check_attached(self.area[k2], self.area[k], h=h) &
                    check_dir(self.area[k2], self.area[k], omode)):
                ans.append(k2)
                ans = self.find_nexeones(k2, mode, ans)
        return ans

    def check_corner_hit(self, area, x, y, w, h):
        delta = 5

        box = self._area2box(area)
        boxd = [box[0]*w, box[1]*h, box[2]*w, box[3]*h]
        xd = x*w
        yd = y*h

        if abs(boxd[0]-xd) < delta:
            if abs(boxd[1]-yd) < delta:
                return 0
            if abs(boxd[3]-yd) < delta:
                return 1
        if abs(boxd[2]-xd) < delta:
            if abs(boxd[1]-yd) < delta:
                return 2
            if abs(boxd[3]-yd) < delta:
                return 3
        return -1

    def check_edge_hit(self, area, event, w, h):
        '''
        check if event hit on edge
        '''
        xd = event.x
        yd = event.y
        area2 = self.apply_pm_2_area(area)
        box = self._area2box(area2)
        boxd = [box[0]*w, box[1]*h, box[2]*w, box[3]*h]
        if (xd < 15) | (w-xd < 15) | (yd < 15) | (h-yd < 15):
            delta = 20
        else:
            delta = 5

        if abs(boxd[0]-xd) < delta:
            if (boxd[1]-yd)*(boxd[3]-yd) < 0:
                event.x = boxd[0]
                return 0
        if abs(boxd[1]-yd) < delta:
            if (boxd[0]-xd)*(boxd[2]-xd) < 0:
                event.y = boxd[1]
                return 1
        if abs(boxd[2]-xd) < delta:
            if (boxd[1]-yd)*(boxd[3]-yd) < 0:
                event.x = boxd[2]
                return 2
        if abs(boxd[3]-yd) < delta:
            if (boxd[0]-xd)*(boxd[2]-xd) < 0:
                event.y = boxd[3]
                return 3
        return -1

    def check_edge_hit2(self, area, xd, yd, w, h):
        '''
        check if event hit on edge or extended
        '''

        area2 = self.apply_pm_2_area(area)
        box = self._area2box(area2)
        boxd = [box[0]*w, box[1]*h, box[2]*w, box[3]*h]
        if (xd < 15) | (w-xd < 15) | (yd < 15) | (h-yd < 15):
            delta = 20
        else:
            delta = 3
        if abs(boxd[0]-xd) < delta:
            return 0
        if abs(boxd[1]-yd) < delta:
            return 1
        if abs(boxd[2]-xd) < delta:
            return 2
        if abs(boxd[3]-yd) < delta:
            return 3
        return -1

    def apply_pm_2_area(self, area):
        if self.nomargin:
            pm = [0., 0., 0., 0.]
        else:
            pm = self.page_margin
        area2 = [(1.-pm[0]-pm[1])*area[0]+pm[0],
                 (1.-pm[2]-pm[3])*area[1]+pm[3],
                 area[2]*(1.-pm[0]-pm[1]),
                 area[3]*(1.-pm[2]-pm[3]), ]
        return area2

    def unapply_pm_2_area(self, area):
        if self.nomargin:
            pm = [0., 0., 0., 0.]
        else:
            pm = self.page_margin
        return [(area[0]-pm[0])/(1.-pm[0]-pm[1]),
                (area[1]-pm[3])/(1.-pm[2]-pm[3]),
                area[2]/(1.-pm[0]-pm[1]),
                area[3]/(1.-pm[2]-pm[3]), ]

    def select_connected_edge(self, single=False):
        '''
        chose all edge among hit2 which are 
        connected to hit
        '''
        area = self.area[self.edge_hit[0][0]]
        #area = self.apply_pm_2_area(area)
        if self.edge_hit[0][1] == 0:
            x = area[0]
        if self.edge_hit[0][1] == 1:
            y = area[1]
        if self.edge_hit[0][1] == 2:
            x = area[0]+area[2]
        if self.edge_hit[0][1] == 3:
            y = area[1]+area[3]

        if single:
            self.edge_hit2 = [tmp[:] for tmp in self.edge_hit]

        # extend edge_hit flag (left edge of the area)
        cx, cy = self.canvas.GetClientSize()
        if self.edge_hit[0][1] == 0 or self.edge_hit[0][1] == 2:
            maxy0, miny0 = self.mark_hit2_edge(0, 1, 3)
            maxy1, miny1 = self.mark_hit2_edge(2, 1, 3)
#           print maxy0, miny0, maxy1, miny1
            if maxy0 == -1 and miny0 == -1:
                return 0, miny1, maxy1, x, 1
            if maxy1 == -1 and miny1 == -1:
                return 0, miny0, maxy0, x, 1
            if (int(maxy0*cy) == int(maxy1*cy) and
                    int(miny0*cy) == int(miny1*cy)):
                return 0, miny0, maxy0, x, 0
        elif self.edge_hit[0][1] == 1 or self.edge_hit[0][1] == 3:
            maxx0, minx0 = self.mark_hit2_edge(1, 0, 2)
            maxx1, minx1 = self.mark_hit2_edge(3, 0, 2)
            if maxx0 == -1 and minx0 == -1:
                return 1, minx1, maxx1, y, 1
            if maxx1 == -1 and minx1 == -1:
                return 1, minx0, maxx0, y, 1
            if (int(maxx0*cx) == int(maxx1*cx) and
                    int(minx0*cx) == int(minx1*cx)):
                return 1, minx0, maxx0, y, 0

        return -1, 0, 0, 0

    def mark_hit2_edge(self, k, m1, m2):
        delta = 5
        cx, cy = self.canvas.GetClientSize()
        if (k == 1) | (k == 3):
            cx = cy
        hit = self.edge_hit
        hit2 = self.edge_hit2
        maxx = -1
        minx = -1
        for i in range(len(hit)):
            if hit[i][1] == k:
                a = self.area[hit[i][0]]
#            a = self.apply_pm_2_area(a)
                maxx = max([a[m1], a[m1]+a[m2]])
                minx = min([a[m1], a[m1]+a[m2]])
        check = True
        while check:
            check = False
            for i in range(len(hit2)):
                if hit2[i][1] == k:
                    if hit2[i][2]:
                        continue
                    a = self.area[hit2[i][0]]
                    if (abs(a[m1] - maxx)*cx < delta or
                        abs(a[m1]+a[m2] - maxx)*cx < delta or
                        abs(a[m1] - minx)*cx < delta or
                            abs(a[m1]+a[m2] - minx)*cx < delta):
                        hit2[i][2] = True
                        check = True
                        maxx = max([maxx, a[m1], a[m1]+a[m2]])
                        minx = min([minx, a[m1], a[m1]+a[m2]])
        return maxx, minx

    def distH(self, area_hit):
        # searching area
        areas = self._search_connecting_area(area_hit, True)
        # adjusting area
        w = sum([a1[2] for a1 in areas])
        w = w/len(areas)
        tmp = sorted([(a1[0], a1) for a1 in areas],  key=lambda x:x[0])
        x0 = tmp[0][0]
        i = 0
        for void, a1 in tmp:
            idx = self.area.index(a1)
            self.area[idx][2] = w
            self.area[idx][0] = x0 + w*i
            i = i+1
        # finish up
        request = [('m', k, a) for k, a in enumerate(self.area)]
        self._finish_up_area_edit(request=request)

    def distV(self, area_hit):
        # searching area
        areas = self._search_connecting_area(area_hit, False)
        # adjusting area
        h = sum([a1[3] for a1 in areas])
        h = h/len(areas)
        tmp = sorted([(a1[1], a1) for a1 in areas],  key=lambda x:x[0])
        y0 = tmp[0][0]
        i = 0
        for void, a1 in tmp:
            k = self.area.index(a1)
            self.area[k][3] = h
            self.area[k][1] = y0 + h*i
            i = i+1
        # finish up
        request = [('m', k, a) for k, a in enumerate(self.area)]
        self._finish_up_area_edit(request=request)

    def apply_left_only(self, area_hit):
        fig = self.figure
        f_page = fig.figobj
        if self.area.count(area_hit) == 0:
            return
        areas = self._search_connecting_area(area_hit, True)

        areas = [b[1] for b in
                 sorted([(a[0], a) for a in areas],  key=lambda x:x[0])]

        f_axes_arr, ac = self._checkup_edge_only_leftbottom(areas, 0)
        new_len = self._calc_axis_edge_only_newlen_bottomleft(
            f_axes_arr, 0, 1, 2)

        # apply new length
        request = []
        p0 = f_axes_arr[0].getp("area")[0]
        fig_page = f_axes_arr[0].get_figpage()
        for j in range(len(f_axes_arr)):
            f_axes = f_axes_arr[j]
            a = f_axes.get_area()
            a[0] = p0
            a[2] = new_len[j, 0]
            request.append(('m', fig_page.get_iaxes(f_axes), a))
            p0 = p0 + new_len[j, 0]

        # finish up
        self._finish_up_area_edit(request=request, ac=ac, name='left only')

    def apply_bottom_only(self, area_hit):

        if self.area.count(area_hit) == 0:
            return
        areas = self._search_connecting_area(area_hit, False)

        # sort area
        areas = [b[1] for b in
                 sorted([(a[1], a) for a in areas],  key=lambda x:x[0])]

        # set _edge_only
        f_axes_arr, ac = self._checkup_edge_only_leftbottom(areas, 1)

        new_len = self._calc_axis_edge_only_newlen_bottomleft(
            f_axes_arr, 3, 2, 3)
        # apply new length
        request = []
        p0 = f_axes_arr[0].getp("area")[1]
        fig_page = f_axes_arr[0].get_figpage()
        for j in range(len(f_axes_arr)):
            f_axes = f_axes_arr[j]
            a = f_axes.get_area()
            a[1] = p0
            a[3] = new_len[j, 0]
            request.append(('m', fig_page.get_iaxes(f_axes), a))
            p0 = p0 + new_len[j, 0]

        # finish up
        self._finish_up_area_edit(request=request, ac=ac, name='bottom only')

    def _calc_axis_edge_only_newlen_bottomleft(self, f_axes_arr, p1, p2, p3):
        '''
        p1 = 3, p2 = 2, p3=3 : bottom
        p1 = 0, p2 = 1, p3=2 : left
        '''
        # find the length of each section
        l = len(f_axes_arr)  # size of problem
        # (1) reset extra margin
        for j in range(l):
            f_axes_arr[j]._extra_margin[p1] = 0.0
        # (2) set extra margin for edge
        text_space = 0.05
        f_axes_arr[0]._extra_margin[p1] = text_space
        # (3) fill matrix
        import numpy as np
        M = np.zeros((l, l))
        for j in range(l-1):
            rect, defm, margin1 = f_axes_arr[j].calc_rect()
            rect, defm, margin2 = f_axes_arr[j+1].calc_rect()
            emargin1 = f_axes_arr[j]._extra_margin
            emargin2 = f_axes_arr[j+1]._extra_margin
            M[j, j] = (1 - margin1[p1] - margin1[p2]
                       - emargin1[p1] - emargin1[p2])
            M[j, j+1] = -(1 - margin2[p1] - margin2[p2]
                          - emargin2[p1] - emargin2[p2])
        d = 0
        for j in range(l):
            M[-1, j] = 1
            d = d + f_axes_arr[j].getp("area")[p3]
        a = np.zeros((l, 1))
        a[-1, 0] = d
        # (4) solve
        new_len = inv(M).dot(a)
        return new_len

    def _checkup_edge_only_leftbottom(self, areas, idx):
        '''
        idx  = 1 : bottom
        idx  = 0 : left
        '''
        fig = self.figure
        f_page = fig.figobj
        f_axes_arr = []
        ac = []
        for f_axes in f_page.walk_axes():
            if f_axes.getp("area") in areas:
                edge_only = f_axes.get_edge_only()
                if f_axes.getp("area") != areas[0]:
                    edge_only[idx] = True
                    ac.append(UndoRedoFigobjMethod(f_axes._artists[0],
                                                   'edge_only', edge_only))
                else:
                    edge_only[idx] = False
                    ac.append(UndoRedoFigobjMethod(f_axes._artists[0],
                                                   'edge_only', edge_only))
                f_axes_arr.append(f_axes)
        f_axes_arr = [b[1] for b in
                      sorted([(a.getp("area")[idx], a) for a in f_axes_arr], key=lambda x:x[0])]
        return f_axes_arr, ac

    def _finish_up_area_edit(self, request=None, ac=None, name='area'):
        self.SetCanvasValue(request=request, ac=ac, name=name)
        wx.CallAfter(self.SetEditorValue)
#       self.canvas.draw()

    def _search_connecting_area(self, area_hit, Flag):
        '''
        Flag = True  : horizontal connection
        Flag = Flase : vertical connection
        '''
        areas = [area_hit]
        hit = True
        while hit:
            hit = False
            for a1 in areas:
                for a2 in self.area:
                    if (not a2 in areas and
                            self._check_attached(a1, a2, Flag)):
                        areas.append(a2)
                        hit = True
        return areas

    def _check_attached(self, a1, a2, h=True):
        def check(a, b):
            return abs(a-b) < 3
        x1, y1, x2, y2 = self._area2boxv(a1)
        x10, y10, x20, y20 = self._area2boxv(a2)
        if (h and (check(y1, y10) and check(y2, y20)) and
                (check(x1, x20) or check(x2, x10))):
               return True
        if (not h and (check(x1, x10) and check(x2, x20)) and
                (check(y1, y20) or check(y2, y10))):
               return True

        return False

    def _search_nearby(self, a):
        x10, y10, x20, y20 = self._area2boxv(a)

        def check(a, b):
            return abs(a-b) < 3

        left = []  # area at the left of a
        right = []
        up = []
        down = []
        for k, a1 in enumerate(self.area):
            x1, y1, x2, y2 = self._area2boxv(a1)
            if check(x10, x2):
                left.append((k, a1[:]))
            elif check(x20, x1):
                right.append((k, a1[:]))
            elif check(y10, y2):
                down.append((k, a1[:]))
            elif check(y20, y1):
                up.append((k, a1[:]))
        return left, right, up, down

    def _area2xy(self, sec):
        # convert x0, y0, w, h to x, y
        x = [sec[0], sec[0]+sec[2], sec[0]+sec[2], sec[0],  sec[0]]
        y = [sec[1], sec[1], sec[1]+sec[3], sec[1]+sec[3], sec[1]]
        return x, y

    def _area2box(self, sec):
        # convert x0, y0, w, h to x0, y0, x1, y1
        return [sec[0], sec[1], sec[0]+sec[2], sec[1]+sec[3]]

    def _area2boxv(self, sec):
        h = self.figure.get_figheight()*self.figure.get_dpi()
        w = self.figure.get_figwidth()*self.figure.get_dpi()
        # convert x0, y0, w, h to x0, y0, x1, y1
        return [sec[0]*w, sec[1]*h,
                (sec[0]+sec[2])*w, (sec[1]+sec[3])*h]

    def _box2area(self, box):
        return [min([box[0], box[2]]), min([box[1], box[3]]),
                abs(box[0]-box[2]), abs(box[1]-box[3])]

    def _load_canvas_value(self):
        ifig_canvas = self.canvas
        if ifig_canvas is None:
            # if property edtior is not linked...
            return

        fig = self.figure
        f_page = fig.figobj

        self.area = []
        self.rect = []
        self.use_def_margin = []
        self.margin = []
        self.def_margin = f_page.getp("def_margin")
        self.page_margin = f_page._page_margin[:]
        self.nomargin = f_page.get_nomargin()
#       print  f_page.get_children()
        for name, f_axes in f_page.get_children():
            if not isinstance(f_axes, FigAxes):
                continue
            area = f_axes.get_area()
            self.area.append(area)
            r1, c1, m1 = f_axes.calc_rect(ignore_pagemargin=True)
            self.rect.append(r1)
            self.margin.append(m1)
            self.use_def_margin.append(c1)
        if self.area_hit == None:
            if len(self.area) != 0:
                self.area_hit = self.area[0]
#       self.draw()

    def SetCanvasValue(self, axes=None, request=None, ac=None, name='area'):
        ifig_canvas = self.canvas
        if request is None:
            ifig_canvas.set_area(self.area)
            ifig_canvas.draw()
        else:
            send_area_request(ifig_canvas, axes=axes,
                              request=request, ac=ac, name=name)
#       ifig_canvas.draw()

    def SetEditorValue(self, axes=None):
        self._load_canvas_value()



from __future__ import print_function

from .gl_compound import GLCompound
from matplotlib.collections import LineCollection
from matplotlib.cm import ScalarMappable
from matplotlib.lines import Line2D

from matplotlib.artist import allow_rasterization
from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.mto.axis_user import XUser, YUser, ZUser, CUser
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
import logging
from ifigure.utils.cbook import ProcessKeywords, LoadImageFile, isdynamic, isiterable
from ifigure.utils.args_parser import ArgsParser
from matplotlib.path import Path
from matplotlib.transforms import Bbox, TransformedPath
from matplotlib.colors import Colormap

from ifigure.matplotlib_mod.cz_linecollection import CZLineCollection
from ifigure.matplotlib_mod.art3d_gl import Line3DCollectionGL, LineGL

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigPlot')

_decimate_limit = 2500
_decimate_limit_switch = 3


class FigPlot(FigObj, XUser, YUser, ZUser, CUser):
    '''
    FigPlot: an object to handle line and errorbar plot.

    See help(plot), help(errorbar) for more detail of
    generating this object from ifigure.interactive

    '''
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._mpl_cmd = 'plot'
            obj._eb_container = (None, tuple(), tuple())
            obj._other_artists = []
            obj._objs = []  # for debug....
            obj._data_extent = None
            obj._is_decimate = False
            obj._use_decimate = False
            obj._decimate_limit = _decimate_limit
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj
        mpl_cmd = kywds.pop('mpl_command', 'plot')
        p = ArgsParser()
        p.add_opt('x', None, ['numbers|nonstr', 'dynamic'])
        p.add_var('y', ['numbers|nonstr', 'dynamic'])
        p.add_opt('z', None, ['numbers|nonstr', 'dynamic'])
        p.add_opt('s', '', 'str')  # optional argument
        p.add_key('cz', False, 'bool')

#        if mpl_cmd is 'errorbar':
        p.add_key('xerr', None)
        p.add_key('yerr', None)
        p.add_key('ecolor', None)
        p.add_key('elinewidth', None)
        p.add_key('array_idx', None)
        if 'cz' in kywds and kywds['cz']:
            p.add_key('cmap', 'jet')
        p.add_key('c',  None)

        p.set_ndconvert("x", "y", "z", "xerr", "yerr", "c")
        p.set_squeeze_minimum_1D("x", "y", "z", "xerr", "yerr", "c")
        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)
        obj._mpl_cmd = mpl_cmd
        if (v["x"] is None and not isdynamic(v["y"])):
            v["x"] = np.arange(v["y"].shape[-1]).astype(v["y"].dtype)
        if 'cmap' in v and v['cmap'] is not None:
            if isinstance(v['cmap'], Colormap):
                v['cmap'] = v['cmap'].name
            kywds['cmap'] = v['cmap']
            del v['cmap']
        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        return obj

    def __init__(self, *args,  **kywds):
        XUser.__init__(self)
        YUser.__init__(self)
        ZUser.__init__(self)
        CUser.__init__(self)

        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigPlot, self).__init__(*args, **kywds)

    @classmethod
    def isFigPlot(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return True

    @classmethod
    def get_namebase(self):
        return 'plot'

    def property_in_file(self):
        #        if isinstance(self._artists[0],
        #                      CZLineCollection):
        #            return (["linewidth", "linestyle", "alpha"] +
        #                    super(FigPlot, self).property_in_file())
        #        else:
        props = (["linestyle", "linewidth", "marker", "markersize",
                  "markeredgecolor", "markerfacecolor",
                  "markeredgewidth", "alpha"] +
                 super(FigPlot, self).property_in_file())
        if not self.getp('cz'):
            props = ['color'] + props
        return props

    @classmethod
    def property_in_palette(self):
        return (["line", "marker", "errorbar"],
                [["color", "linestyle", "linewidth_2", "alpha"],
                 ["markerfacecolor", "markeredgecolor", "marker",
                  "markersize", "markeredgewidth"],
                 ["ecolor", "elinewidth"]],)

    @classmethod
    def attr_in_file(self):
        return ['cz', 'ecolor', 'elinewidth']+super(FigPlot, self).attr_in_file()

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'plot1.png')
        return [idx1]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)
        ZUser.unset_az(self)
        CUser.unset_ac(self)

        super(FigPlot, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)
        ZUser.get_zaxisparam(self)
        CUser.get_caxisparam(self)

    def args2var(self):
        names0 = self.attr_in_file()
        names = ["x", "y", "z", "s", "c"]
        use_np = [True]*3 + [False] + [True, ]
        if self._mpl_cmd != 'plot':
            names = names + ["xerr", "yerr"]
            use_np = use_np + [True]*2
        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)
        x = values[0]
        y = values[1]
        if x is None:
            x = np.arange((y.shape)[-1]).astype(y.dtype)
            self.setp("x", x)
        if x is None:
            return False
        if y is None:
            return False
        if (x.ndim == 1 and y.ndim == 1 and
                x.size != y.size):
            self.setp("x", None)
            self.setp("y", None)
            return False
        if self._mpl_cmd != 'plot':
            xerr, yerr = self.getp(('xerr', 'yerr'))
            if xerr is not None and xerr.size == 1:
                self.setp('xerr',
                          np.array([xerr[0]]*(self.getp("y").size)))
            if yerr is not None and yerr.size == 1:
                self.setp('yerr',
                          np.array([yerr[0]]*(self.getp("y").size)))
        if 'cmap' in self.getvar("kywds"):
            cax = self.get_caxisparam()
            cax.set_cmap(self.getvar('kywds')['cmap'])

#        self.setp("x", self.getp('x').astype(np.float64))
#        self.setp("y", self.getp('y').astype(np.float64))
        return True

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change
        #import inspect
        # print [(x[1],x[3]) for x in inspect.stack()]
        # print 'generating plot', self, self._suppress
        #import traceback
        # traceback.print_stack()
        container = self.get_container()
        if container is None:
            return
        if self.isempty() is False:
            return

        if self._mpl_cmd == 'plot':
            x, y, z, s, c = self._eval_xy()  # this handles "use_var"
        else:
            x, y, z, s, c, xerr, yerr = self._eval_xy()  # this handles "use_var"

        lp = self.getp("loaded_property")
        cz = self.getp('cz')
        axes = self.get_figaxes()


#           if lp is None:
        if True:
            #              x, y, z, s, c = self.getp(("x", "y", "z", "s", "c"))
            if y is None:
                return
            if x is None:
                return

            if (x is not None and
                    y is not None):
                if len(y.shape) == 1:
                    kywds = self._var["kywds"].copy()
                    if self._mpl_cmd == 'plot':
                        if not cz:
                            if 'cmap' in kywds:
                                del kywds['cmap']
                            if (x.size > self._decimate_limit and
                                self._use_decimate and
                                    self.check_data_uniform(x)):
                                self.set_artist(container.plot([0], [0],
                                                               s, **kywds))
                                self._is_decimate = True
                                self.handle_axes_change()
                            else:
                                if np.iscomplexobj(x):
                                    x = np.array(x, copy=False).real
                                if np.iscomplexobj(y):
                                    y = np.array(y, copy=False).real
                                args = [x, y]
                                self._is_decimate = False
                                axes = self.get_figaxes()
                                if axes.get_3d():
                                    if z is None:
                                        z = np.array([0]*len(x))
                                    else:
                                        if np.iscomplexobj(z):
                                            z = np.array(z, copy=False).real
                                    args.append(z)
                                    kywds['array_idx'] = self.getvar(
                                        'array_idx')
                                args.append(s)
                                self.set_artist(container.plot(*args,
                                                               **kywds))
                                if axes.get_3d():
                                    self = convert_to_FigPlotGL(self)
#                             self._artists[0].set_lod(True)
                        else:
                            cax = self.get_caxisparam()
                            cmap = cax.get_cmap()
                            if cmap is None:
                                cmap = 'jet'
                                cax.set_cmap(cmap)
                            args = [x, y, z]
                            if c is not None and axes.get_3d():
                                args.append(c)

                            a = container.cz_plot(*args,  **kywds)
                            self.set_artist(a)

                    elif self._mpl_cmd == 'errorbar':
                        if axes.get_3d():
                            return
                        #xerr, yerr = self.getp(("xerr", "yerr"))
                        kywds['fmt'] = s
                        a = container.errorbar(x,
                                               y,
                                               xerr=xerr,
                                               yerr=yerr,
                                               **kywds)
                        self.set_artist(a[0])
                        self._eb_container = a
                        if self.getp('ecolor') is None:
                            self.setp('ecolor', self.get_ecolor(a[0]))
                        else:
                            self.set_ecolor(self.getp('ecolor'), a[0])

                        if self.getp('elinewidth') is None:
                            self.setp('elinewidth',
                                      self.get_elinewidth(a[0]))
                        else:
                            self.set_elinewidth(self.getp('elinewidth'),
                                                a[0])
        if lp is not None:
            for i in range(0, len(lp)):
                self.set_artist_property(self._artists[i], lp[i])
            self.delp("loaded_property")

        self.set_rasterized()

        mappable = self.get_mappable()
        ac = self.get_caxisparam()
        for a in mappable:
            ac.set_crangeparam_to_artist(a)

#           for artist in self._artists:
#               artist.figobj=self
#               artist.figobj_hl=[]
#               artist.set_zorder(self.getp('zorder'))
#               self._objs.append(weakref.ref(artist))

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = artist

        self.store_loaded_property()

        if len(artistlist) is len(self._artists):
            for name, child in self.get_children():
                child.del_artist(delall=True)

        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            for a in artistlist:
                #             a.set_picker(None)
                try:
                    if a in container.collections:
                        a.remove()
                    elif a in container.lines:
                        a.remove()
                    if self._mpl_cmd == 'errorbar':
                        for a in self._eb_container[1]:
                            a.remove()
                        for a in self._eb_container[2]:
                            a.remove()
                        self._eb_container = (None, tuple(), tuple())
                except:
                    logging.exception("fig_plot: highlight_aritst() failed")
            for a in self._other_artists:
                a.remove()

        super(FigPlot, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist

        container = self.get_container()
        if val == True:
            for a in alist:
                if isinstance(a, LineGL):
                    hl = alist[0].add_hl_mask()
                elif isinstance(a, Line2D) or isinstance(a, CZLineCollection):
                    if hasattr(a, 'set_3d_properties'):
                        args = a._verts3d
                    else:
                        x = a.get_xdata()
                        y = a.get_ydata()
                        p = 300
                        if x.shape[0] > p:
                            idx = np.floor(
                                np.arange(p)*(x.shape[0]/float(p))).astype(int)
                        else:
                            idx = np.arange(x.shape[0])
                        args = [x[idx].copy(), y[idx].copy()]
                    kwargs = {'marker': 's',
                              'color': 'k',
                              'linestyle': 'None',
                              'markerfacecolor': 'None',
                              'markeredgewidth': 0.5,
                              'scalex': False, 'scaley': False}

                    hl = container.plot(*args, **kwargs)

#                  if hasattr(a, 'set_3d_properties'):
#                      z=a._verts3d[2]
#                      hl[0].set_3d_properties(zs = z)
                elif isinstance(a, Line3DCollectionGL):
                    hl = alist[0].add_hl_mask()
#                  hl = a.make_hl_artist(container) #slower
                for item in hl:
                    a.figobj_hl.append(item)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    hl.remove()
                a.figobj_hl = []

    def get_mappable(self):
        return [a for a in self._artists if isinstance(a, ScalarMappable)]

#
#   def hit_test
#
    def picker_a(self, artist, evt):
        axes = artist.axes
        if axes is None:
            return False, {}
        hit, extra = artist.contains(evt)
        if hit:
            return True, {"child_artist": artist}
        else:
            return False, {}

    def picker_a0(self, artist, evt):
        hit, extra = self.picker_a(artist, evt)
        if hit:
            return hit, extra, 'area', 3
        else:
            return False, {}, None, 0
#
#  Popup in Canvas
#

    def canvas_menu(self):
        ac = self.get_caxisparam()
        if len(ac._member) > 1:
            pass
        cz = self.getp('cz')
        if ac is None:
            return super(FigPlot, self).tree_viewer_menu()
        if cz and ac.cbar_shown():
            m = [('Hide ColorBar', self.onHideCB1, None), ]
        elif cz and not ac.cbar_shown():
            m = [('Show ColorBar', self.onShowCB1, None), ]
        else:
            m = []
        menus = m + \
            super(FigPlot, self).canvas_menu()
        return menus

    def get_export_val(self, a):
        from matplotlib.artist import getp
        data = {"xdata": self.getvar("x"),
                "ydata": self.getvar("y")}
        zdata = self.getvar("z")
        cdata = self.getvar("c")
        if cdata is not None:
            data["cdata"] = cdata
            if zdata is not None:
                data["zdata"] = zdata
        else:
            if zdata is not None:
                if self.getvar("cz"):
                    data["cdata"] = zdata
                else:
                    data["zdata"] = zdata
        return data

    def get_data_extent(self):
        #        print 'entering data extent',  self._data_extent
        if self._data_extent is not None:
            return self._data_extent
        if self._mpl_cmd == 'plot':
            x, y, z, s, c = self._eval_xy()  # this handles "use_var"
        else:
            x, y, z, s, c, xerr, yerr = self._eval_xy()  # this handles "use_var"

#        if self.isempty():
        if x is None:
            self._data_extent = [0, len(y), np.nanmin(y), np.nanmax(y)]
        else:
            self._data_extent = [np.nanmin(x), np.nanmax(x),
                                 np.nanmin(y), np.nanmax(y)]
#        else:
#            xr = (np.inf, -np.inf)
#            yr = (np.inf, -np.inf)
#            for a in self._artists:
#                x = a.get_xdata()
#                y = a.get_ydata()
#                xr = (min((xr[0], np.nanmin(x))), max((xr[1], np.nanmax(x))))
#                yr = (min((yr[0], np.nanmin(y))), max((yr[1], np.nanmax(y))))
#            self._data_extent=[xr[0], xr[1], yr[0], yr[1]]
#        print 'exiting data extent',  self._data_extent
        return self._data_extent
#
#   error bar properties
#

    def set_ecolor(self, value, a=None):
        if self._mpl_cmd == 'errorbar':
            self.setp('ecolor', value)
            for a in self._eb_container[1]:
                a.set_color(value)
            for a in self._eb_container[2]:
                a.set_color(value)

    def get_ecolor(self, a):
        v = None
        if self._mpl_cmd == 'errorbar':
            for a in self._eb_container[1]:
                v = a.get_color()
            for a in self._eb_container[2]:
                v = a.get_color()
            if isinstance(v, np.ndarray):
                v = np.squeeze(v)
            return v
        else:
            return v

    def set_elinewidth(self, value, a=None):
        if self._mpl_cmd == 'errorbar':
            self.setp('elinewidth', value)
            for a in self._eb_container[1]:
                a.set_linewidth(value)
            for a in self._eb_container[2]:
                a.set_linewidth(value)

    def get_elinewidth(self, a):
        v = None
        if self._mpl_cmd == 'errorbar':
            for a in self._eb_container[1]:
                v = a.get_linewidth()
            for a in self._eb_container[2]:
                v = a.get_linewidth()
            if isiterable(v):
                v = v[0]
            return v
        else:
            return v

    def set_linewidth(self, value, a=None):
        a.set_linewidth(value)

    def get_linewidth(self, a):
        try:
            return a.get_linewidth()[0]
        except:
            return a.get_linewidth()

    def set_decimate(self, value, a):
        if (self._mpl_cmd == 'plot' and
                not self.getp('cz')):
            self._use_decimate = value
            if self._use_decimate:
                self.handle_axes_change()
            else:
                x = self.getp('x')
                y = self.getp('y')
                self._artists[0].set_xdata(x)
                self._artists[0].set_ydata(y)
                if hasattr(self._artists[0], 'set_czdata'):
                    z = self.getp('z')
                    self._artists[0].set_czdata(z)
        else:
            self._use_decimate = False

    def get_decimate(self, a):
        return self._use_decimate

    def check_data_uniform(self, x):
        if any(np.isnan(x)):
            self._use_decimate = False
            return False
        skip = x.size//10
        ix = range(10)
        t = [x[ix*skip] - x[(ix+1)*skip] for ix in range(9)]
        # print t
        # print abs((max(t) - min(t))/np.mean(t))
        if abs((max(t) - min(t))/np.mean(t)) > 0.01:
            self._use_decimate = False
        return abs((max(t) - min(t))/np.mean(t)) < 0.01
#
#   range
#

    def get_xrange(self, xrange=[None, None], scale='linear'):
        #        de = self.get_data_extent()[0:2]
        #        if de is not None:
        if self._mpl_cmd == 'plot':
            x, y, z, s, c = self._eval_xy()  # this handles "use_var"
        else:
            x, y, z, s, c, xerr, yerr = self._eval_xy()  # this handles "use_var"
        if x is None:
            return xrange
        if scale == 'log':
            x = mask_negative(x)
        return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        def calc_yrange(x, y, xrange, yrange, scale):
            # x.ndim = 1, y.ndim = 1 in this func
            xidx = np.isfinite(x)
            idx = (np.where((x[xidx] >= xrange[0]) &
                            (x[xidx] <= xrange[1])))
            check = False
            for ar in idx:
                if len(ar) == 0:
                    check = True
            if check:
                return yrange

            yy = y[xidx][idx]
            if scale == 'log':
                yy = mask_negative(yy)
            y0n = np.nanmin(yy)
            y1n = np.nanmax(yy)
            yrange = self._update_range(yrange, (y0n, y1n))
            return yrange

        de = self.get_data_extent()
        y0 = yrange[0]
        y1 = yrange[1]

        if (xrange[0] is not None and
                xrange[1] is not None):
            if self._mpl_cmd == 'plot':
                x, y, z, s, c = self._eval_xy()
            else:
                x, y, z, s, c, xerr, yerr = self._eval_xy()

            if x.ndim > 2 or y.ndim > 2:
                return yrange
            if x.ndim == 1:
                xl = 1
            else:
                xl = x.shape[0]
            if y.ndim == 1:
                yl = 1
            else:
                yl = y.shape[0]
            for k in range(xl):
                if x.ndim == 1:
                    xt = x
                else:
                    xt = x[k]
                for m in range(yl):
                    if y.ndim == 1:
                        yt = y
                    else:
                        yt = y[m]
                    yrange = calc_yrange(xt, yt, xrange, yrange, scale)

        if (not (xrange[0] is not None and
                 xrange[1] is not None) or
                (yrange[0] is None) or (yrange[1] is None)):
            if de is not None:
                yrange = self._update_range(yrange, de[2:])
        return yrange

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None], scale='linear'):
        if self._mpl_cmd == 'plot':
            x, y, z, s, c = self._eval_xy()
        else:
            x, y, z, s, c, xerr, yerr = self._eval_xy()
        if c is not None:
            if scale == 'log':
                c = mask_negative(c)
            crange = self._update_range(crange,
                                        (np.nanmin(c), np.nanmax(c)))
        else:
            if z is not None and self.getp('cz'):
                if scale == 'log':
                    z = mask_negative(z)
                crange = self._update_range(crange,
                                            (np.nanmin(z), np.nanmax(z)))
        return crange

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None], scale='linear'):
        if self._mpl_cmd == 'plot':
            x, y, z, s, c = self._eval_xy()
        else:
            x, y, z, s, c, xerr, yerr = self._eval_xy()
        if z is not None:
            if scale == 'log':
                z = mask_negative(z)
            zrange = self._update_range(zrange, (np.nanmin(z), np.nanmax(z)))
        return zrange

    def _saveload_names(self):
        return ['x', 'y', 'z', 'xerr', 'yerr', 's', 'c']

    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        names = self._saveload_names()
        var = {name: check(self, name) for name in names}
        for name in names:
            if not var[name]:
                if self._save_mode == 0 or name == 's':
                    var[name+'data'] = self.getp(name)
                else:
                    if self.getp(name) is not None:
                        var[name+'data'] = np.array([0, 0])
                    else:
                        var[name+'data'] = None
        var["mpl_cmd"] = self._mpl_cmd
        var["use_decimate"] = self._use_decimate
        data['FigPlot'] = (1, var)
        data = super(FigPlot, self).save_data2(data)

        return data

    def prepare_compact_savemode(self):
        var_bk = self._var.copy()
        self._var['x'] = np.array([0, 0])
        self._var['y'] = np.array([0, 0])
        self._var['c'] = None
        self._var['z'] = None
        self._var['xerr'] = None
        self._var['yerr'] = None

        return var_bk

    def load_data2(self, data):
        d = data['FigPlot']
        super(FigPlot, self).load_data2(data)
        var = d[1]
        self._mpl_cmd = var["mpl_cmd"]
        if "use_decimate" in var:
            self._use_decimate = var["use_decimate"]
        names = self._saveload_names()
        for name in names:
            if not name in var:
                var[name] = True
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def refresh_artist_data(self):
        if self._mpl_cmd != 'plot':
            FigObj.refresh_artist_data(self)
            return
        if self.isempty() and not self._suppress:
            self.generate_artist()
            return
        self._data_extent = None
        if self._mpl_cmd == 'plot':
            x, y, z, s, c = self._eval_xy()

        if x.ndim == 0:
            x = np.array([x])
        if y.ndim == 0:
            y = np.array([y])
        self._artists[0].set_xdata(x)
        self._artists[0].set_ydata(y)
        self.handle_axes_change()

    def handle_axes_change(self, evt=None):
        '''
        decimation when the range changed
        '''
#        print ' handle_axes_change(self,evt)'
        if False:  # skip this part...
            if (self._mpl_cmd == 'plot' and
                    len(self._artists) != 0):
                x = self.getp('x')
                y = self.getp('y')
                yrange = self.get_yaxisparam().range
                xrange = self.get_xaxisparam().range
                idx = self._get_idx_from_range(x, xrange)

                self._artists[0].set_xdata(x[idx])
                self._artists[0].set_ydata(y[idx])
                if hasattr(self._artists[0], 'set_czdata'):
                    z = self.getp('z')
                    self._artists[0].set_czdata(z[idx])
            else:
                pass
        x = self.getp('x')
        y = self.getp('y')
        if (self._use_decimate and
                x.size > self._decimate_limit*_decimate_limit_switch):
            yrange = self.get_yaxisparam().range
            xrange = self.get_xaxisparam().range
            idx = self._get_idx_from_range(x, xrange)

            x2, y2 = self._decimate_array_method2(x, y, idx)
            x2 = np.concatenate(
                ([x[max((0, idx[0]-1))]], x2, [x[min((len(x)-1, idx[-1]+1))]]))
            y2 = np.concatenate(
                ([y[max((0, idx[0]-1))]], y2, [y[min((len(y)-1, idx[-1]+1))]]))
            self._artists[0].set_xdata(x2)
            self._artists[0].set_ydata(y2)
            self._is_decimate = True
        else:
            self._is_decimate = False
        mappable = self.get_mappable()
        ac = self.get_caxisparam()
        for a in mappable:
            ac.set_crangeparam_to_artist(a)

    def _eval_xy(self):
        if self._mpl_cmd == 'plot':
            names = ("x", "y", "z", "s", "c")
        else:
            names = ("x", "y", "z", "s", "c", "xerr", "yerr")

        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return [None]*len(names)
        return self.getp(names)

    def _get_idx_from_range(self, x, range):
        idx = np.where((x[np.isfinite(x)] > range[0]) &
                       (x[np.isfinite(x)] < range[1]))[0]
        if idx.size > 0:
            if idx[0] != 0:
                idx = np.hstack((idx[0]-1, idx))
            if idx[-1] != x.size-1:
                idx = np.hstack((idx, idx[-1]+1))
            return idx
        else:
            return np.array([np.argmin(np.abs(x-range[0]))])

    def _decimate_array_method2(self, x, y, idx):
        x2 = x[idx]
        y2 = y[idx]

        chunksize = self._decimate_limit//2
        numchunks = y2.size // chunksize
        if numchunks == 0:
            # when size of y2 is too small..
            return x2, y2

        if y2.size//numchunks > chunksize:
            chunksize = y2.size//numchunks

        ychunks_f = y2[:chunksize*numchunks].reshape((-1, numchunks))
        xchunks_f = x2[:chunksize*numchunks].reshape((-1, numchunks))

        y3 = np.vstack((np.nanmin(ychunks_f, axis=1),
                        np.nanmax(ychunks_f, axis=1))).transpose().reshape(chunksize*2)
        x3 = np.vstack((np.nanmin(xchunks_f, axis=1),
                        np.nanmax(xchunks_f, axis=1))).transpose().reshape(chunksize*2)
        # append left over..
        if chunksize*numchunks < y2.size:
            x2 = np.hstack(
                (x3, np.nanmin(x2[chunksize*numchunks:]), np.nanmax(x2[chunksize*numchunks:])))
            y2 = np.hstack(
                (y3, np.nanmin(y2[chunksize*numchunks:]), np.nanmax(y2[chunksize*numchunks:])))
            return x2, y2
        else:
            return x3, y3

    def set_alpha(self, v, a):
        a.set_alpha(v)


class StepPlot(FigPlot):
    '''
    stepplot is a special case of figplot.
    if has only x and y data. 
    it supports decimation
    it supports data compression using file.

    Because of historiy, routines for decimation is 
    written in FigPLot
    '''
    def __new__(cls, *args, **kywds):
        if 'src' in kywds:
            obj = FigPlot.__new__(cls, *args, **kywds)
#            obj = set_hidden_vars(obj)
            obj._use_decimate = False
            return obj
        mpl_cmd, kywds = ProcessKeywords(kywds, 'mpl_command', 'plot')
        mpl_cmd = 'plot'
        p = ArgsParser()
        p.add_opt('x', None, ['numbers|nonstr', 'dynamic'])
        p.add_var('y', ['numbers|nonstr', 'dynamic'])
        p.add_opt('s', '', 'str')  # optional argument

#        if mpl_cmd is 'errorbar':

        p.set_ndconvert("x", "y")
        p.set_squeeze_minimum_1D("x", "y")

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigPlot.__new__(cls, *args, **kywds)
#        obj = set_hidden_vars(obj)
        obj._use_decimate = False
        obj._mpl_cmd = mpl_cmd
        if (v["x"] is None and not isdynamic(v["y"])):
            v["x"] = np.arange(v["y"].shape[-1])
        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        return obj

    @classmethod
    def get_namebase(self):
        return 'stepplot'

    def get_export_val(self, a):
        self.expand_catalog()
        data = {"xdata": self.getvar('x'),
                "ydata": self.getvar('y')}
        return data

    def _eval_xy(self):
        names = ("x", "y", "z", "s", "c")
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return [None]*len(names)

            self._expand_stepplot_data()
        return self.getp(names)

    def _expand_stepplot_data(self):
        x = self.getp('x')
        y = self.getp('y')

        dx = np.hstack(((x[1:]-x[:-1])/1e6, 0))
        x = np.transpose(np.vstack((x, x+dx))).flatten()[:-1]
        y = np.transpose(np.vstack((y, y))).flatten()[:-1]
        self.setp('x', np.array(x))
        self.setp('y', np.array(y))

        #new_x = [x[0]]
        #new_y = [y[0]]
        # for k in range(x.size-1):
        #  new_x.append(x[k+1] -  (x[k+1] - x[k])/1e6)
        #  new_x.append(x[k+1])
        #  new_y.append(y[k])
        #  new_y.append(y[k+1])
        # print new_x
        # print new_y


class TimeTrace(FigPlot):
    '''
    time trace is a special case of figplot.
    if has only x and y data. 
    it supports decimation
    it supports data compression using file.

    Because of historiy, routines for decimation is 
    written in FigPLot

    '''
    def __new__(cls, *args, **kywds):
        if 'src' in kywds:
            obj = FigPlot.__new__(cls, *args, **kywds)
#            obj = set_hidden_vars(obj)
            obj._use_decimate = True
            return obj
        mpl_cmd, kywds = ProcessKeywords(kywds, 'mpl_command', 'plot')
        mpl_cmd = 'plot'
        p = ArgsParser()
        p.add_opt('x', None, ['numbers|nonstr', 'dynamic'])
        p.add_var('y', ['numbers|nonstr', 'dynamic'])
        p.add_opt('s', '', 'str')  # optional argument

#        if mpl_cmd is 'errorbar':

        p.set_ndconvert("x", "y")
        p.set_squeeze_minimum_1D("x", "y")

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigPlot.__new__(cls, *args, **kywds)
#        obj = set_hidden_vars(obj)
        obj._use_decimate = True
        obj._mpl_cmd = mpl_cmd
        if (v["x"] is None and not isdynamic(v["y"])):
            v["x"] = np.arange(v["y"].shape[-1])
        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        return obj

    @classmethod
    def property_in_palette(self):
        tab, names = super(TimeTrace, self).property_in_palette()
        names[0].append('decimate')
        return tab[:2], names[:2]

    @classmethod
    def get_namebase(self):
        return 'timetrace'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'timetrace.png')
        return [idx1]

    def refresh_artist_data(self):
        if self.isempty() and not self._suppress:
            self.generate_artist()
            return
        self._data_extent = None

        x, y, z, s, c = self._eval_xy()  # this handles "use_var"

        if x.ndim == 0:
            x = np.array([x])
        if y.ndim == 0:
            y = np.array([y])
        self.check_data_uniform(x)  # this could make _use_decimate False

        self.handle_axes_change()
#        dprint1('refresh_artist_data :'+
#                str(self._artists[0].get_xdata().size) + ' ' +
#                str(self._artists[0].get_ydata().size))
#        self._artists[0].set_ydata(y)
#        self._artists[0].set_xdata(x)

    def expand_catalog(self):
        if not self.hasvar('x_catalog'):
            return
        try:
            x = self.getvar('x_catalog').restore(0)
        except IOError:
            print("IOError during catalog expansion")
            self.delvar('x_catalog')
            self.delvar('y_catalog')
            return
        try:
            y = self.getvar('y_catalog').restore(0)
        except IOError:
            print("IOError during catalog expansion")
            self.delvar('x_catalog')
            self.delvar('y_catalog')
            return

        self.delvar('x_catalog')
        self.delvar('y_catalog')
        if x.size != y.size:
            dprint1('catalog data x and y has different length')
            l = min([x.size, y.size])
            x = x[:l]
            y = y[:l]
        self.setvar('x', x)
        self.setvar('y', y)
        self.setp('x', x)
        self.setp('y', y)
        return x, y

    def get_export_val(self, a):
        self.expand_catalog()
        return {"xdata": self.getvar('x'),
                "ydata": self.getvar('y')}

    def handle_axes_change(self, evt=None):
        '''
        decimation when the range changed
        '''
#        print ' handle_axes_change(self,evt)'
        if len(self._artists) != 0:
            x = self.getp('x')
            y = self.getp('y')
            yrange = self.get_yaxisparam().range
            xrange = self.get_xaxisparam().range
#                print xrange, yrange
            idx = self._get_idx_from_range(x, xrange)
            if ((idx.size < self._decimate_limit) and
                    self.hasvar('x_catalog') and
                    self.hasvar('y_catalog')):
                dprint1('expanding high resolution data ' +
                        self.get_full_path())
                x, y = self.expand_catalog()
                idx = self._get_idx_from_range(x, xrange)
            if (self._use_decimate and
                    idx.size > self._decimate_limit*_decimate_limit_switch):
                x2, y2 = self._decimate_array_method2(x, y, idx)
                x2 = np.concatenate(
                    ([x[max((0, idx[0]-1))]], x2, [x[min((len(x)-1, idx[-1]+1))]]))
                y2 = np.concatenate(
                    ([y[max((0, idx[0]-1))]], y2, [y[min((len(y)-1, idx[-1]+1))]]))
                self._artists[0].set_xdata(x2)
                self._artists[0].set_ydata(y2)
                self._is_decimate = True
            else:
                self._artists[0].set_xdata(x[idx])
                self._artists[0].set_ydata(y[idx])
                self._is_decimate = False
    #                dprint1(str(time.time()-t))
        else:
            pass
        mappable = self.get_mappable()
        ac = self.get_caxisparam()
        for a in mappable:
            ac.set_crangeparam_to_artist(a)

    def save_data2(self, data=None):
        return FigPlot.save_data2(self, data)

    def load_data2(self, data):
        return FigPlot.load_data2(self, data)


class FigPlotGL(GLCompound, FigPlot):
    def onSetRotCenter(self, evt):
        x, y, z = self.getvar('x', 'y', 'z')

        sel = self.getSelectedIndex()
        if len(sel) == 0:
            sel = self.shown_component

        array_idx = self.getvar('array_idx')

        if array_idx is not None and len(sel) > 0:
            idx = np.in1d(array_idx, sel)
            vv = np.vstack((x[idx], y[idx], z[idx])).transpose()
            cc = np.mean(vv, 0)
        else:
            cc = np.array([np.mean(x),
                           np.mean(y),
                           np.mean(z)])

        axes = self._artists[0].axes
        axes._gl_rot_center = cc
        axes._gl_use_rot_center = True
        evt.Skip()

    def canvas_menu(self):
        m = FigObj.canvas_menu(self)
        m2 = GLCompound.canvas_menu(self)
        return m[:1]+m2+m[1:]

def convert_to_FigPlotGL(obj):
    obj.__class__ = FigPlotGL
    obj.set_pickmask(obj._pickmask)
    return obj



from __future__ import print_function
import numpy as np
import sys


def XY2pts(X, Y):
    if not isinstance(X, np.ndarray):
        X = np.array(X)
    if not isinstance(Y, np.ndarray):
        Y = np.array(Y)

    pts = np.hstack((X.flatten()[:, np.newaxis],
                     Y.flatten()[:, np.newaxis]))
    return pts


def ensure_cyclic(pts):
    if pts[0, 0] != pts[-1, 0] or pts[0, 1] != pts[-1, 1]:
        return np.vstack((pts, pts[0]))
    return pts


def norm_path(x):
    return np.sqrt(np.sum(x**2, 1))


def dot_path(x, y):
    return np.sum(x*y, 1)


def cross_path(x, y):
    d = x*np.fliplr(y)
    return d[:, 0] - d[:, 1]


def path_len(path):
    d = path[1:]-path[:-1]
    dd = np.sqrt(np.sum(d**2, 1))
    return np.sum(dd)


def check_inside(rec, x, y):
    ret = True

    if rec[0] > x:
        ret = False
    if rec[0]+rec[2] < x:
        ret = False
    if rec[1] > y:
        ret = False
    if rec[1]+rec[3] < y:
        ret = False
    return ret


def check_boxoverwrap(box1, box2):
    xo = False
    yo = False

    if ((box1[0] < box2[0] and
         box1[1] > box2[0]) or
        (box1[0] < box2[1] and
         box1[1] > box2[1])):
        xo = True
    if ((box1[2] < box2[2] and
         box1[3] > box2[2]) or
        (box1[2] < box2[3] and
         box1[3] > box2[3])):
        yo = True

    return xo and yo


def calc_scale_r(rec, old_rect):
    scale = [(rec[1]-rec[0])/(old_rect[1]-old_rect[0]),
             0.,
             0.,
             (rec[3]-rec[2])/(old_rect[3]-old_rect[2]),
             rec[0]-old_rect[0],
             rec[2]-old_rect[2]]
    return scale


def calc_scale(rec, old_rect):

    scale = [(rec[1]-rec[0])/(old_rect[1]-old_rect[0]),
             0.,
             0.,
             (rec[3]-rec[2])/(old_rect[3]-old_rect[2])]

    if old_rect[1] == old_rect[0]:
        scale[0] = 1
    if old_rect[3] == old_rect[2]:
        scale[3] = 1
    scale = scale + [rec[0] - scale[0]*old_rect[0],
                     rec[2] - scale[3]*old_rect[2]]
    return scale


def scale_rect(rec, scale):
    # this should be called scale area
    # rect is [x0, x1, y0, y1]
    #  [scale[0], scale[1]] x + scale[4]
    #  [scale[2], scale[3]] y + scale[5]
    return [scale[0]*rec[0] + scale[1]*rec[2] + scale[4],
            scale[0]*rec[1] + scale[1]*rec[3] + scale[4],
            scale[2]*rec[0] + scale[3]*rec[2] + scale[5],
            scale[2]*rec[1] + scale[3]*rec[3] + scale[5], ]


def scale_rect_r(rec, scale):
    # this should be called scale area
    # rect is [x0, x1, y0, y1]
    dx = rec[1]-rec[0]
    dy = rec[3]-rec[2]
    return [rec[0]+scale[4],
            rec[0]+dx*scale[0]+scale[4],
            rec[2]+scale[5],
            rec[2]+dy*scale[3]+scale[5]]


def transform_point(t, x, y):
    v = t.transform(np.array([[x, y]]))
    return v[0, 0], v[0, 1]


def area_intersection(p1, p2, area, internal_only=False,
                      closed=True, once=False, return_idx=False):
    '''
    search all possible crosspoint with area on the line connecting p1 and p2
    if internal_only, crosspoint should be between p1 and p2
    p1, p2 : 2D points (x, y)
    area   : [[x1, x2, x3, x4,,,,], [y1, y2, y3,,,,]]
    '''
    x_a = area[0]
    y_a = area[1]
    if closed:
        if x_a[0] != x_a[-1] or y_a[0] != y_a[-1]:
            y_a.append(y_a[0])
            x_a.append(x_a[0])

    p1 = np.array(p1)
    p2 = np.array(p2)

    p = p2 - p1

    x = []
    y = []
    idx = []
    q1 = XY2pts(x_a[:-1], y_a[:-1])
    q2 = XY2pts(x_a[1:],  y_a[1:])
    q0 = q2 - q1
    for k in range(len(area[0])-1):
        #      q1 = np.array([x_a[k], y_a[k]])
        #      q2 = np.array([x_a[k+1], y_a[k+1]])
        #      q  = q2 - q1
        q = q0[k]
        m = np.matrix((-p, q))
        if np.linalg.cond(m) < 1./sys.float_info.epsilon:
            #          ans =  np.array((p1 - q1)*(np.matrix((-p, q))**-1)).flatten()
            ans = np.array((p1 - q1[k])*(np.matrix((-p, q))**-1)).flatten()
        else:
            continue

        if 0. < ans[-1] <= 1.:
            if not internal_only or (internal_only and
                                     (0. < ans[0] <= 1.)):
                #              x.append((q1 + ans[1]*q)[0])
                #              y.append((q1 + ans[1]*q)[1])
                x.append((q1[k] + ans[1]*q)[0])
                y.append((q1[k] + ans[1]*q)[1])
                idx.append(k+ans[1])
                if once:
                    break

    if return_idx:
        return x, y, idx
    else:
        return x, y


def area_intersection_x(x1, x2, y0, area, internal_only=False,
                        once=False, debug=False, return_idx=False):
    '''
    search all possible crosspoint with area on the line connecting p1 and p2
    if internal_only, crosspoint should be between p1 and p2
    p1, p2 : 2D points (x, y)
    area   : [[x1, x2, x3, x4,,,,], [y1, y2, y3,,,,]]
    '''
    x_a = area[0]
    y_a = area[1]

    p1 = np.array([x1, y0])
    p2 = np.array([x2, y0])

    p = p2 - p1

    x = []
    y = []
    idx = []
    for k in range(len(area[0])-1):
        if (y_a[k]-y0)*(y_a[k+1]-y0) > 0:
            continue
        q1 = np.array([x_a[k], y_a[k]])
        q2 = np.array([x_a[k+1], y_a[k+1]])
        q = q2 - q1
        m = np.matrix((-p, q))
        if np.linalg.cond(m) < 1./sys.float_info.epsilon:
            ans = np.array((p1 - q1)*(np.matrix((-p, q))**-1)).flatten()
        else:
            continue
        if debug:
            print(q1, q2, ans)
        if 0. < ans[-1] <= 1.:
            if not internal_only or (internal_only and
                                     (0. < ans[0] <= 1.)):
                x.append((q1 + ans[1]*q)[0])
                y.append((q1 + ans[1]*q)[1])
                idx.append(k + ans[1])
                if once:
                    break
    if return_idx:
        return x, y, idx
    else:
        return x, y


def area_intersection_y(x0, y1, y2, area, internal_only=False,
                        once=False, debug=False, return_idx=False):
    '''
    search all possible crosspoint with area on the line connecting p1 and p2
    if internal_only, crosspoint should be between p1 and p2
    p1, p2 : 2D points (x, y)
    area   : [[x1, x2, x3, x4,,,,], [y1, y2, y3,,,,]]
    '''
    x_a = area[0]
    y_a = area[1]
    p1 = np.array([x0, y1])
    p2 = np.array([x0, y2])

    p = p2 - p1

    x = []
    y = []
    idx = []
    for k in range(len(area[0])-1):
        if (x_a[k] - x0)*(x_a[k+1] - x0) > 0:
            continue
        q1 = np.array([x_a[k], y_a[k]])
        q2 = np.array([x_a[k+1], y_a[k+1]])
        q = q2 - q1
        m = np.matrix((-p, q))
        if np.linalg.cond(m) < 1./sys.float_info.epsilon:
            ans = np.array((p1 - q1)*(np.matrix((-p, q))**-1)).flatten()
        else:
            continue
        if debug:
            print(ans)
        if 0. < ans[-1] <= 1.:
            if not internal_only or (internal_only and
                                     (0. < ans[0] <= 1.)):
                x.append((q1 + ans[1]*q)[0])
                y.append((q1 + ans[1]*q)[1])
                idx.append(k + ans[1])
                if once:
                    break
    if return_idx:
        return x, y, idx
    else:
        return x, y


def path_cut(path, p1, p2):
    '''
    cut path by a line connecting p1 and p2
    '''
    p1 = np.array(p1)
    p2 = np.array(p2)
    p = p2 - p1

    x = []
    y = []
    for k in range(len(path)-1):
        q1 = path[k]
        q2 = path[k+1]
        q = q2 - q1
        m = np.matrix((-p, q))
        if np.linalg.cond(m) < 1./sys.float_info.epsilon:
            ans = np.array((p1 - q1)*(np.matrix((-p, q))**-1)).flatten()
        else:
            continue

        if (0 < ans[-1] <= 1) and (0 < ans[0] <= 1):
            xc, yc = q1 + ans[1]*q
            break
    else:
        return [path, ]
    return [np.vstack((path[:k, :], np.array((xc, yc)))),
            np.vstack((np.array((xc, yc)), path[k:, :]))]


def path_contain(path, xy, check=False, cyclic=True):
    '''
    check if path contains xy based on winding number
    algorithm.

    equivalent to 
      mod(angle(u)-angle(v) - 180, 360) - 180 ???

    if xy is on a point of path, it returns True

    the first point and the last point should be the same
    (ensure_cyclic forces it)a
    '''
    if cyclic:
        path = ensure_cyclic(path)
    import numpy as np
    dx = path[:, 0] - xy[0]
    dy = path[:, 1] - xy[1]
    d1 = np.sqrt(dx[:-1]**2 + dy[:-1]**2)
    d2 = np.sqrt(dx[1:]**2 + dy[1:]**2)

    d = (dx[:-1]*dy[1:] - dx[1:]*dy[:-1])/d1/d2
    dotp = (dx[:-1]*dx[1:] + dy[1:]*dy[:-1])
    if any(np.isnan(d)):
        # when d1, d2 = 0
        return True

    d = np.arcsin(d)
    d[np.logical_and(dotp < 0, d > 0)] = np.pi - \
        d[np.logical_and(dotp < 0, d > 0)]
    d[np.logical_and(dotp < 0, d < 0)] = -np.pi - \
        d[np.logical_and(dotp < 0, d < 0)]

    xxxx = sum(d)
#    print(np.abs(xxxx)/3.14159/2)
#    if check: return d
    if check:
        return np.abs(xxxx)/np.pi/2
    return -0.001 < np.abs(xxxx)-np.pi*2 < 0.001


def calc_area(path, xy=[0., 0.]):
    '''
    calc area inside path
    '''
    import numpy as np
    dx = path[:, 0] - xy[0]
    dy = path[:, 1] - xy[1]
    d1 = np.sqrt(dx[:-1]**2 + dy[:-1]**2)
    d2 = np.sqrt(dx[1:]**2 + dy[1:]**2)

    d = (dx[:-1]*dy[1:] - dx[1:]*dy[:-1])
    return np.abs(np.sum(d)/2.)


def dist_path2xy(path, xy, check=False, return_idx=False):
    '''
    find distance from path to xy and
    the closest point on the path
    '''
    dx = path[:, 0] - xy[0]
    dy = path[:, 1] - xy[1]

    dd = XY2pts(dx, dy)
    dpath = path[1:] - path[:-1]

    #  b cos() = vec(a)*vec(b)/norm(a)
    ddcos = dot_path(-dd[:-1], dpath)/norm_path(dpath)
    norm_dpath = norm_path(dpath)
    norm_dd = norm_path(dd)
    idx = ddcos/norm_dpath

    #  b sin() = vec(a)xvec(b)/norm(a) (this is distance)
    ddsin = np.abs(cross_path(-dd[:-1], dpath)/norm_dpath)
    for k in range(len(idx)):
        if idx[k] < 0.:
            idx[k] = 0
            ddsin[k] = norm_dd[k]
        if idx[k] > 1.:
            idx[k] = 1
            ddsin[k] = norm_dd[k+1]
#      if ddcos[k] < 0:
#           idx[k] = 0.
#           ddsin[k] = norm_dd[k]
#      elif ddcos[k] > norm_dpath[k]:
#           idx[k] = 1.
#           ddsin[k] = norm_dd[k+1]
    min_k = np.argmin(ddsin)
    x, y = path[min_k] + (path[min_k+1]-path[min_k])*idx[min_k]

    if return_idx:
        return ddsin[min_k], x, y, min_k + idx[min_k]
    else:
        return ddsin[min_k], x, y


def s_path(path, idx=None):
    '''
    return distance along the path
    if idx is give, it returns the s at idx. idx
    can be float.
    '''
    norm_dpath = norm_path(path[1:] - path[:-1])
    cumsum = np.cumsum(norm_dpath)
    if idx is None:
        return cumsum
    else:
        lidx = int(idx)
        didx = idx - lidx
        return cumsum[lidx] + norm_dpath[lidx] * didx


'''
def connect_pairs(ll):
    d = {}

    flags = [False,]*len(ll)
    count = 1
    d[ll[0][0]] = ll[0][1]
    flags[0] = True
    while count < len(ll):
       for k, f in enumerate(flags):
           if f: continue
           l = ll[k]
           if l[0] in d.keys():
               d[l[1]] = l[0]
               flags[k] = True
               count = count + 1
           elif l[1] in d.keys():
               d[l[0]] = l[1]
               flags[k] = True
               count = count + 1               
           else:
               pass
    key = d.keys()[0]
    pt = [key]
    lmax = len(d.keys())
    while d[key] != pt[0]:
        pt.append(d[key])
        key = d[key]
        if len(pt) > lmax: break
    if d[key] == pt[0]: pt.append(pt[0])

    return pt
'''


def connect_pairs(ll):
    '''
    connect paris of indices to make a loop

    (example)
    >>> idx = array([[1, 4],  [3, 4], [1,2], [2, 3],])
    >>> connect_pairs(idx)
    [[1, 2, 3, 4, 1]]
    '''
    if not isinstance(ll, np.ndarray):
        ll = np.array(ll)

    idx = np.where(ll[:, 0] > ll[:, 1])[0]
    t1 = ll[idx, 0]
    t2 = ll[idx, 1]
    ll[idx, 0] = t2
    ll[idx, 1] = t1

    ii = np.vstack([np.arange(ll.shape[0]), ]*2).transpose()
    d = np.ones(ll.shape[0]*ll.shape[1]).reshape(ll.shape)
    from scipy.sparse import csr_matrix, coo_matrix
    m = coo_matrix((d.flatten(), (ii.flatten(), ll.flatten())),
                   shape=(len(ll), np.max(ll+1)))
    mm = m.tocsc()
    ic = mm.indices
    icp = mm.indptr
    mm = m.tocsr()
    ir = mm.indices
    irp = mm.indptr

    def get_start(taken):
        idx = np.where(np.logical_and(np.diff(icp) == 1, taken == 0))[0]
        nz = np.where(np.logical_and(np.diff(icp) != 0, taken == 0))[0]
        if len(nz) == 0:
            return
        if len(idx) > 0:
            #print('Open end found')
            pt = (ic[icp[idx[0]]], idx[0])
        else:
            pt = (ic[icp[nz[0]]], nz[0])
        pts = [pt]
        return pts

    def hop_v(pt):
        ii = pt[1]
        ii = [icp[ii], icp[ii+1]-1]
        next = ic[ii[1]] if ic[ii[0]] == pt[0] else ic[ii[0]]
        return (next, pt[1])

    def hop_h(pt):
        ii = pt[0]
        ii = [irp[ii], irp[ii+1]-1]
        next = ir[ii[1]] if ir[ii[0]] == pt[1] else ir[ii[0]]
        return (pt[0], next)

    def trace(pts):
        loop = [pts[-1][1]]
        while True:
            pts.append(hop_v(pts[-1]))
            # rows.append(pts[-1][0])
            pts.append(hop_h(pts[-1]))
            if pts[-1][1] in loop:
                break  # open ended
            loop.append(pts[-1][1])
            if pts[-1] == pts[0]:
                break
        return loop

    taken = (icp*0)[:-1]
    loops = []
    while True:
        pts = get_start(taken)
        if pts is None:
            break
        loop = trace(pts)
        loops.append(loop)
        taken[loop] = 1

    return loops


def make_loop_idx(loop):
    ''' 
    make indexset array from loop

    (example)
    a = connect_pairs(idx)
    figure()
    solid(v,  make_loop_idx(a[0]))
    '''
    if loop[0] == loop[-1]:
        loop = loop[:-1]
    return np.vstack((loop, np.roll(loop, 1))).transpose()


def find_edges(idxset):
    '''
    find exterior edges of triangulation, but the algorith
    is not limited to triangles.

    idxset is an array defining patch.
    for example :idxset[:, 3] is idx of triangles
    this routine findds edges which appears once
    patch

    '''
    s = idxset.shape
    l = []  # loop
    for i in range(s[-1]-1):
        l.append(idx[:, [i, i+1]])
    l.append(idx[:, [s[-1]-1, 0]])
    c = np.vstack(l)
    from collections import defaultdict
    d = defaultdict(int)
    for i in c:
        d[tuple(np.sort(i))] += 1
    e = [k for k in d if d[k] == 1]

    return e
